'''
The HTTP server to serve local portal and provide RESTful APIs
'''

import os
import pkg_resources
from twisted.web.server import Site
from twisted.web.static import File
from twisted.internet import reactor
from twisted.internet.endpoints import serverFromString
from klein import Klein
from autobahn.twisted.resource import WebSocketResource

from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.core.system.system_status import SystemStatus

from .airshark_api import AirsharkApi
from .airshark_ws import AirsharkSpectrumFactory, AirsharkAnalyzerFactory
from .audio_api import AudioApi
from .auth import requires_auth, AuthApi
from .change_api import ChangeApi
from .change_ws import ChangeStreamFactory
from .chute_api import ChuteApi
from .chute_log_ws import ChuteLogWsFactory
from .config_api import ConfigApi
from .information_api import InformationApi
from .log_sockjs import LogSockJSFactory
from .network_api import NetworkApi
from .paradrop_log_ws import ParadropLogWsFactory
from .password_api import PasswordApi
from .password_manager import PasswordManager
from .snapd_resource import SnapdResource
from .status_sockjs import StatusSockJSFactory
from .token_manager import TokenManager


class HttpServer(object):
    app = Klein()

    def __init__(self, update_manager, update_fetcher, airshark_manager, portal_dir=None):
        self.update_manager = update_manager
        self.update_fetcher = update_fetcher
        self.system_status = SystemStatus()
        self.password_manager = PasswordManager()
        self.token_manager = TokenManager()
        self.airshark_manager = airshark_manager

        if portal_dir:
            self.portal_dir = portal_dir
        elif 'SNAP' in os.environ:
            self.portal_dir = os.environ['SNAP'] + '/www'
        else:
            self.portal_dir = pkg_resources.resource_filename('paradrop', 'static')


    @app.route('/api/v1/audio', branch=True)
    def api_audio(self, request):
        return AudioApi().routes.resource()


    @app.route('/api/v1/auth', branch=True)
    def api_auth(self, request):
        return AuthApi(self.password_manager, self.token_manager).routes.resource()


    @app.route('/api/v1/info', branch=True)
    @requires_auth
    def api_information(self, request):
        return InformationApi().routes.resource()


    @app.route('/api/v1/changes/', branch=True)
    @requires_auth
    def api_changes(self, request):
        return ChangeApi(self.update_manager).routes.resource()


    @app.route('/api/v1/config', branch=True)
    @requires_auth
    def api_configuration(self, request):
        return ConfigApi(self.update_manager, self.update_fetcher).routes.resource()


    @app.route('/api/v1/chutes/', branch=True)
    @requires_auth
    def api_chute(self, request):
        return ChuteApi(self.update_manager).routes.resource()


    @app.route('/api/v1/password', branch=True)
    @requires_auth
    def api_password(self, request):
        return PasswordApi(self.password_manager).routes.resource()


    @app.route('/api/v1/airshark', branch=True)
    @requires_auth
    def api_airshark(self, request):
        return AirsharkApi(self.airshark_manager).routes.resource()


    @app.route('/api/v1/network', branch=True)
    @requires_auth
    def api_network(self, request):
        return NetworkApi().routes.resource()


    @app.route('/snapd/', branch=True)
    @requires_auth
    def snapd(self, request):
        return SnapdResource()


    '''
    # Not being used for now because we are using HAProxy to setup the reverse proxy
    @app.route('/chutes/<string:name>', branch=True)
    def chutes(self, request, name):
        try:
            container = ChuteContainer(name)
            ip = container.getIP()
            return ReverseProxyResource(ip, 80, '/')
        except ParadropException as error:
            return str(error)
    '''

    @app.route('/sockjs/logs/<string:name>', branch=True)
    @requires_auth
    def logs(self, request, name):
        #cors.config_cors(request)
        chute = ChuteStorage.get_chute(name)
        factory = LogSockJSFactory(chute)
        factory.setProtocolOptions(autoPingInterval=5, autoPingTimeout=2)
        return WebSocketResource(factory)


    @app.route('/sockjs/status', branch=True)
    @requires_auth
    def status(self, request):
        #cors.config_cors(request)
        factory = StatusSockJSFactory(self.system_status)
        factory.setProtocolOptions(autoPingInterval=5, autoPingTimeout=2)
        return WebSocketResource(factory)


    @app.route('/ws/chute_logs/<string:name>', branch=True)
    @requires_auth
    def chute_logs(self, request, name):
        #cors.config_cors(request)
        chute = ChuteStorage.get_chute(name)
        factory = ChuteLogWsFactory(chute)
        factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
        return WebSocketResource(factory)


    @app.route('/ws/airshark/analyzer', branch=True)
    @requires_auth
    def airshark_analyzer(self, request):
        #cors.config_cors(request)
        factory = AirsharkAnalyzerFactory(self.airshark_manager)
        factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
        return WebSocketResource(factory)


    @app.route('/ws/airshark/spectrum', branch=True)
    @requires_auth
    def airshark_spectrum(self, request):
        #cors.config_cors(request)
        factory = AirsharkSpectrumFactory(self.airshark_manager)
        factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
        return WebSocketResource(factory)

    @app.route('/ws/paradrop_logs', branch=True)
    @requires_auth
    def paradrop_logs(self, request):
        #cors.config_cors(request)
        factory = ParadropLogWsFactory()
        factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
        return WebSocketResource(factory)

    @app.route('/ws/changes/<int:change_id>/stream', branch=True)
    @requires_auth
    def change_stream(self, request, change_id):
        change = self.update_manager.find_change(change_id)
        if change is not None:
            factory = ChangeStreamFactory(change)
            factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
            return WebSocketResource(factory)
        else:
            request.setResponseCode(404)
            return "{}"


    @app.route('/', branch=True)
    @requires_auth
    def home(self, request):
        return File(self.portal_dir)


def setup_http_server(http_server, host, port):
    endpoint = serverFromString(
        reactor,
        "tcp:port={0}:interface={1}".format(port, host)
    )
    endpoint.listen(Site(http_server.app.resource()))


def annotate_routes(router, prefix):
    """
    Annotate klein routes for compatibility with autoflask generator.
    """
    router.static_path = ""
    router.view_functions = router.endpoints
    for rule in router.url_map.iter_rules():
        rule.doc_prefix = prefix


annotate_routes(AudioApi.routes, "/api/v1/audio")
annotate_routes(ChuteApi.routes, "/api/v1/chutes")
annotate_routes(ConfigApi.routes, "/api/v1/config")
annotate_routes(InformationApi.routes, "/api/v1/info")



###################################################################
# Copyright 2013-2016 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.chute.chute import Chute
from paradrop.core.config import state
from paradrop.core.container import dockerapi

from . import plangraph


def validate_change(update):
    """
    Check if the update is a valid change.
    """
    requires_old = ['stop', 'start', 'delete', 'restart']
    requires_no_old = ['create']

    if update.old is None and update.updateType in requires_old:
        update.failure = "No chute found with id: " + update.name
        return False

    if update.old is not None and update.updateType in requires_no_old:
        update.failure = "Chute already exists."
        return False

    if update.new.state == Chute.STATE_INVALID:
        update.failure = "Chute state is invalid."
        return False

    if update.updateType == "start" and update.old.isRunning():
        update.failure = "Chute is already running."
        return False

    if update.updateType == "stop" and not update.old.isRunning():
        update.failure = "Chute is already stopped."
        return False

    if update.old is None:
        old_version = 0
    else:
        old_version = update.old.version

    if update.updateType == "update" and update.new.version <= old_version:
        update.failure = "Version {} of the chute is already installed.".format(
            old_version)
        return False

    return True


def generatePlans(update):
    """
        This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
        then adds Plan() calls to make the Chute match the @newChute.

        Returns:
            True: abort the plan generation process
    """
    out.verbose("%r\n" % (update))

    # Check for some error conditions that we can detect during the planning
    # stage, e.g. starting a chute that does not exist.
    if not validate_change(update):
        return True

    update.plans.addPlans(plangraph.STATE_GENERATE_SERVICE_PLANS,
                          (generate_service_plans, ))

    if update.new.isRunning():
        update.plans.addPlans(plangraph.STATE_NET_START,
                              (dockerapi.setup_net_interfaces, ),
                              (dockerapi.cleanup_net_interfaces, ))

        if update.old is None:
            update.plans.addPlans(plangraph.STATE_CREATE_BRIDGE,
                                  (dockerapi.create_bridge, ),
                                  (dockerapi.remove_bridge, ))

    if update.old is not None and update.old.isRunning():
        update.plans.addPlans(plangraph.STATE_NET_STOP,
                              (dockerapi.cleanup_net_interfaces, ),
                              (dockerapi.setup_net_interfaces, ))

    if update.updateType == "delete":
        update.plans.addPlans(plangraph.STATE_REMOVE_BRIDGE,
                              (dockerapi.remove_bridge, ),
                              (dockerapi.create_bridge, ))

    # Save the chute to the chutestorage.
    update.plans.addPlans(plangraph.STATE_SAVE_CHUTE, (state.saveChute, ),
                          (state.revertChute, ))

    return None


def generate_service_plans(update):
    """
    Generate plans that depend on the services configured for the chute.

    This needs to happen after the chute configuration has been parsed.
    """
    for service in update.new.get_services():
        if update.updateType in ["create", "update"]:
            update.plans.addPlans(plangraph.STATE_BUILD_IMAGE,
                                  (dockerapi.prepare_image, service),
                                  (dockerapi.remove_image, service))

            update.plans.addPlans(plangraph.STATE_CHECK_IMAGE,
                                  (dockerapi.check_image, service))

        if update.new.isRunning():
            update.plans.addPlans(plangraph.STATE_CALL_START,
                                  (dockerapi.start_container, service),
                                  (dockerapi.remove_container, service))

    old_services = []
    if update.old is not None:
        old_services = update.old.get_services()

    for service in old_services:
        if update.old.isRunning():
            update.plans.addPlans(plangraph.STATE_CALL_STOP,
                                  (dockerapi.remove_container, service),
                                  (dockerapi.start_container, service))

        if update.updateType in ["update", "delete"]:
            update.plans.addPlans(plangraph.STATE_CALL_CLEANUP,
                                  (dockerapi.remove_image, service))



"""
The Chute Builder interprets the chute specification, which comes from a
paradrop.yaml file or from a JSON object via a cloud update. The Chute
Builder then produces a valid Chute ojbect with one or more Service
objects.

There are different versions of the chute specification in code
repositories, and so the intent is to encapsulate this complexity within
the ChuteBuilder implementation. Other code modules, especially the
execution plan pipeline, should be able to rely on a relatively stable
Chute and Service object interface.

+----------------------------+                   +---------------+
|                            |         interpret |               |
| Chute Specification        | <-----------------+ Chute Builder |
| serialized as YAML or JSON |                   |               |
| in a variety of formats    |                   ++-------------++
|                            |                    |             |
+----------------------------+                    |  construct  |
                                                  |             |
                                                  v             |
                                                                |
+----------------+                         +-------+            |
|                |                         |       |            |
| Execution Plan |                         | Chute |            v
| -------------- +-----------------------> |       | has 1+
| function 1     |                         +-------o---+---------+
| ...            | consume                             |         |
| function N     |                                     | Service |
|                +-----------------------------------> |         |
+----------------+                                     +---------+
"""
import six

from .chute import Chute
from .service import Service


WIRELESS_OPTIONS = set([
    "ssid",
    "key",
    "nasid",
    "acct_server",
    "acct_secret",
    "acct_interval",
    "hidden",
    "isolate",
    "maxassoc"
])


def fix_interface_type(interface):
    orig_type = interface.get("type")
    orig_mode = interface.get("mode", "ap")

    if orig_type == "wifi":
        return "wifi-{}".format(orig_mode)
    else:
        return orig_type


def fix_wireless_options(interface):
    wireless = interface.get("wireless", {})

    for key, value in six.iteritems(interface):
        if key in WIRELESS_OPTIONS:
            wireless[key] = value

    for key, value in six.iteritems(interface.get("options", {})):
        if key in WIRELESS_OPTIONS:
            wireless[key] = value

    return wireless


class ChuteBuilder(object):
    """
    Build a composite chute object from a chute specification.

    Implementations of ChuteBuilder are responsible for interpreting
    the chute specification, which comes from a paradrop.yaml file or
    JSON object via cloud update. The ChuteBuilder then produces a valid
    Chute object with one or more Service objects.
    """
    def configure_chute(self, spec):
        for field in ["name", "version", "description"]:
            value = spec.get(field, "unknown")
            setattr(self.chute, field, value)

    def create_chute(self, spec):
        return NotImplemented

    def create_services(self, spec):
        return NotImplemented

    def get_chute(self):
        return self.chute


class SingleServiceChuteBuilder(ChuteBuilder):
    """
    Build a pre-0.12 single-service chute.

    ** Example configuration**:

    .. sourcecode:: yaml

       name: seccam
       description: A Paradrop chute that performs motion detection using a simple WiFi camera.
       version: 1

       net:
         wifi:
           type: wifi
           intfName: wlan0
           dhcp:
             start: 4
             limit: 250
             lease: 12h
           ssid: seccam42
           key: paradropseccam
           options:
             isolate: true
             maxassoc: 100
       web:
         port: 5000
    """

    # Fields that should be present in updates but not chute objects.
    UpdateSpecificArgs = ['deferred']

    def create_chute(self, spec):
        self.chute = Chute()
        self.chute.name = spec.get("name")
        self.chute.description = spec.get("description", None)
        self.chute.environment = spec.get("environment", {})
        self.chute.owner = spec.get("owner", None)
        self.chute.version = spec.get("version", None)

        config = spec.get("config", {})
        self.chute.config = config
        self.chute.web = config.get("web", {})

        # Old chutes did not specify the web service because it was implied.
        if "port" in self.chute.web:
            self.chute.web['service'] = "main"

    def create_services(self, spec):
        service = Service(chute=self.chute, name="main")

        service.command = spec.get("command", None)
        service.image = spec.get("use", None)
        service.source = "."
        service.type = spec.get("type", "normal")

        config = spec.get("config", {})

        # Covert old-style structure with arbitrary names to new-style that is
        # indexed by interface name.
        interfaces = {}
        for iface in config.get("net", {}).values():
            iface['type'] = fix_interface_type(iface)

            if iface['type'].startswith("wifi"):
                iface['wireless'] = fix_wireless_options(iface)

            interfaces[iface['intfName']] = iface

        requests = {}
        requests['as-root'] = spec.get("as_root", False)

        try:
            requests['port-bindings'] = config['host_config']['port_bindings']
        except KeyError:
            requests['port-bindings'] = {}

        # Find extra build options, particularly for light chutes.
        build = {}
        for key in ["image_source", "image_version", "packages"]:
            if key in config:
                build[key] = config[key]

        service.build = build
        service.environment = config.get("environment", {})
        service.interfaces = interfaces
        service.requests = requests

        self.chute.add_service(service)


class MultiServiceChuteBuilder(ChuteBuilder):
    """
    Build a modern multi-service chute.

    ** Example configuration**:

    .. sourcecode:: yaml

       name: seccam
       description: A Paradrop chute that performs motion detection using a simple WiFi camera.
       version: 1

       services:
         main:
           type: light
           source: .
           image: python2
           command: python -u seccam.py

           environment:
             IMAGE_INTERVAL: 2.0
             MOTION_THRESHOLD: 40.0
             SECCAM_MODE: detect

           interfaces:
             wlan0:
               type: wifi-ap

               dhcp:
                 leasetime: 12h
                 limit: 250
                 start: 4

               wireless:
                 ssid: seccam42
                 key: paradropseccam
                 hidden: false
                 isolate: true

               requirements:
                 hwmode: 11g

           requests:
             as-root: true
             port-bindings:
               - external: 81
                 internal: 81

         db:
           type: image
           image: mongo:3.0

       web:
         service: main
         port: 5000
    """
    def create_chute(self, spec):
        self.chute = Chute({})

        for field in ["name", "version", "description", "owner"]:
            value = spec.get(field, "unknown")
            setattr(self.chute, field, value)

        self.chute.environment = spec.get("environment", {})
        self.chute.web = spec.get("web", {})

    def create_services(self, spec):
        for name, spec in six.iteritems(spec.get("services", {})):
            service = Service(chute=self.chute, name=name)

            service.command = spec.get("command", None)
            service.image = spec.get("image", None)
            service.source = spec.get("source", ".")
            service.type = spec.get("type", "normal")

            service.build = spec.get("build", {})
            service.environment = spec.get("environment", {})
            service.interfaces = spec.get("interfaces", {})
            service.requests = spec.get("requests", {})

            # Make sure the intfName field exists for code that depends on it.
            for name, iface in six.iteritems(service.interfaces):
                iface['intfName'] = name

            self.chute.add_service(service)


def build_chute(spec):
    if 'services' in spec:
        builder = MultiServiceChuteBuilder()
    else:
        builder = SingleServiceChuteBuilder()

    builder.create_chute(spec)
    builder.create_services(spec)

    return builder.get_chute()


def rebuild_chute(spec, updates):
    for field in ["name", "version", "description", "owner", "environment", "web"]:
        if field in updates:
            spec[field] = updates[field]

    return build_chute(spec)



import ipaddress
import os

from paradrop.lib.utils import pdosq

from .base import ConfigObject, ConfigOption
from .command import Command, KillCommand, FunctionCommand


class ConfigDhcp(ConfigObject):
    typename = "dhcp"

    options = [
        ConfigOption(name="interface", required=True),
        ConfigOption(name="leasetime"),
        ConfigOption(name="limit", type=int),
        ConfigOption(name="start", type=int),
        ConfigOption(name="dhcp_option", type=list, default=[]),

        # Non-standard options:
        #
        # relay: comma-separated string, specifying a DHCP relay, as with
        #        "<local address>,<server address>[,<interface>]".
        #        Refer to --dhcp-relay in man dnsmasq.
        ConfigOption(name="relay", type=list, default=[])
    ]


class ConfigDomain(ConfigObject):
    typename = "domain"

    options = [
        ConfigOption(name="name", type=str),
        ConfigOption(name="ip", type=str)
    ]

    def getName(self):
        return self.name


class ConfigDnsmasq(ConfigObject):
    typename = "dnsmasq"

    options = [
        ConfigOption(name="authoritative", type=bool, default=True),
        ConfigOption(name="cachesize", type=int, default=150),
        ConfigOption(name="dhcp_boot"),
        ConfigOption(name="dhcpleasemax", type=int, default=1000),
        ConfigOption(name="domain", type=str),
        ConfigOption(name="enable_tftp", type=bool, default=False),
        ConfigOption(name="expandhosts", type=bool, default=True),
        ConfigOption(name="interface", type=list),
        ConfigOption(name="leasefile", type=str),
        ConfigOption(name="noresolv", type=bool, default=False),
        ConfigOption(name="server", type=list),
        ConfigOption(name="tftp_root")
    ]

    def apply(self, allConfigs):
        commands = list()

        # visibleName will be used in choosing file names for this dnsmasq
        # instance, must be unique if there are multiple dnsmasq instances
        visibleName = self.internalName

        if self.interface is None:
            interfaces = []
            for section in self.findByType(allConfigs, "dhcp", "dhcp"):
                interfaces.append(section.interface)
        else:
            interfaces = self.interface

        self.__leasefile = self.leasefile
        if self.__leasefile is None:
            self.__leasefile = "{}/dnsmasq-{}.leases".format(
                self.manager.writeDir, visibleName)
        pdosq.makedirs(os.path.dirname(self.__leasefile))

        pidFile = "{}/dnsmasq-{}.pid".format(
            self.manager.writeDir, visibleName)
        pdosq.makedirs(os.path.dirname(pidFile))

        outputPath = "{}/dnsmasq-{}.conf".format(
            self.manager.writeDir, visibleName)
        pdosq.makedirs(os.path.dirname(outputPath))

        with open(outputPath, "w") as outputFile:
            outputFile.write("#" * 80 + "\n")
            outputFile.write("# dnsmasq configuration file generated by "
                             "pdconfd\n")
            outputFile.write("# Source: {}\n".format(self.source))
            outputFile.write("# Section: {}\n".format(str(self)))
            outputFile.write("#" * 80 + "\n")
            outputFile.write("\n")
            outputFile.write("dhcp-leasefile={}\n".format(self.__leasefile))

            if self.authoritative:
                outputFile.write("dhcp-authoritative\n")
            outputFile.write("cache-size={}\n".format(self.cachesize))
            if self.dhcp_boot is not None:
                outputFile.write("dhcp-boot={}\n".format(self.dhcp_boot))
            outputFile.write("dhcp-lease-max={}\n".format(self.dhcpleasemax))
            if self.domain:
                outputFile.write("domain={}\n".format(self.domain))
            if self.enable_tftp:
                outputFile.write("enable-tftp\n")
            if self.expandhosts:
                outputFile.write("expand-hosts\n")
            if self.noresolv:
                outputFile.write("no-resolv\n")
            if self.tftp_root is not None:
                outputFile.write("tftp-root={}\n".format(self.tftp_root))

            if self.server:
                for server in self.server:
                    outputFile.write("server={}\n".format(server))

            # TODO: Bind interfaces allows us to have multiple instances of
            # dnsmasq running, but it would probably be better to have one
            # running and reconfigure it when we want to add or remove
            # interfaces.  It is not very disruptive to reconfigure and restart
            # dnsmasq.
            outputFile.write("\n")
            outputFile.write("except-interface=lo\n")
            outputFile.write("bind-interfaces\n")

            for intfName in interfaces:
                interface = self.lookup(allConfigs, "network", "interface", intfName)
                outputFile.write("\n")
                outputFile.write("# Options for section interface {}\n".
                                 format(interface.name))
                outputFile.write("interface={}\n".format(
                                 interface.config_ifname))

                network = ipaddress.IPv4Network(u"{}/{}".format(
                    interface.ipaddr, interface.netmask), strict=False)

                dhcp = self.lookup(allConfigs, "dhcp", "dhcp", intfName)

                outputFile.write("\n")
                outputFile.write("# Options for section dhcp {}\n".
                                 format(interface.name))

                if None not in [dhcp.start, dhcp.limit, dhcp.leasetime]:
                    # TODO: Error checking!
                    firstAddress = network.network_address + dhcp.start
                    lastAddress = firstAddress + dhcp.limit

                    outputFile.write("dhcp-range={},{},{},{}\n".format(
                        str(firstAddress), str(lastAddress), interface.netmask,
                        dhcp.leasetime))

                # Write options sections to the config file.
                for option in dhcp.dhcp_option:
                    outputFile.write("dhcp-option={}\n".format(option))

                for relay in dhcp.relay:
                    outputFile.write("dhcp-relay={}\n".format(relay))
                if dhcp.relay:
                    outputFile.write("dhcp-proxy\n")

            outputFile.write("\n")

            for domain in self.findByType(allConfigs, "dhcp", "domain"):
                outputFile.write("address=/{}/{}\n".format(domain.name, domain.ip))

        cmd = ["dnsmasq", "--conf-file={}".format(outputPath),
               "--pid-file={}".format(pidFile)]
        commands.append((self.PRIO_START_DAEMON, Command(cmd, self)))

        self.pidFile = pidFile
        return commands

    def revert(self, allConfigs):
        commands = list()

        commands.append((-self.PRIO_START_DAEMON,
            KillCommand(self.pidFile, self)))

        # Clean up leases and pid files.
        commands.append((-self.PRIO_START_DAEMON,
            FunctionCommand(self, pdosq.safe_remove, self.__leasefile)))
        commands.append((-self.PRIO_START_DAEMON,
            FunctionCommand(self, pdosq.safe_remove, self.pidFile)))

        return commands



'''
The WAMP session of the paradrop daemon
'''
from __future__ import print_function
from twisted.internet.defer import inlineCallbacks
from autobahn.wamp import auth

from paradrop.lib.misc import pdinstall
from paradrop.base.output import out
from paradrop.base import nexus
from paradrop.base.cxbr import BaseSession


def ensure_unicode(s):
    if hasattr(s, 'decode'):
        return s.decode('ascii')
    else:
        return s


class WampSession(BaseSession):
    # Make this a class variable because a new WampSession object is created
    # whenever the WAMP connection resets, but we only call set_update_fetcher
    # on program initialization.
    update_fetcher = None

    def __init__(self, *args, **kwargs):
        self.uriPrefix = 'org.paradrop.'
        super(WampSession, self).__init__(*args, **kwargs)

    def onConnect(self):
        out.info('Router session connected.')
        if (nexus.core.info.pdid and nexus.core.getKey('apitoken')):
            out.info('Starting WAMP-CRA authentication on realm "{}" as user "{}"...'\
                     .format(self.config.realm, nexus.core.info.pdid))
            self.join(self.config.realm, [u'wampcra'],
                    ensure_unicode(nexus.core.info.pdid))

    def onChallenge(self, challenge):
        if challenge.method == u"wampcra":
            out.verbose("WAMP-CRA challenge received: {}".format(challenge))

            wampPassword = nexus.core.getKey('apitoken')
            if u'salt' in challenge.extra:
                # salted secret
                key = auth.derive_key(wampPassword,
                                      challenge.extra['salt'],
                                      challenge.extra['iterations'],
                                      challenge.extra['keylen'])
            else:
                # plain, unsalted secret
                key = wampPassword

            # compute signature for challenge, using the key
            signature = auth.compute_wcs(key, challenge.extra['challenge'])

            # return the signature to the router for verification
            return signature

        else:
            raise Exception("Invalid authmethod {}".format(challenge.method))


    @inlineCallbacks
    def onJoin(self, details):
        out.info('Router session joined')

        yield self.subscribe(self.updatesPending, self.uriPrefix + 'updatesPending')
        yield self.register(self.update, self.uriPrefix + 'update')
        yield BaseSession.onJoin(self, details)


    def onLeave(self, details):
        out.info("Router session left: {}".format(details))
        nexus.core.wamp_connected = False
        self.disconnect()


    def onDisconnect(self):
        out.info("Router session disconnected.")

    def update(self, pdid, data):
        print("Sending command {} to pdinstall".format(data['command']))
        success = pdinstall.sendCommand(data['command'], data)

        # NOTE: If successful, this process will probably be going down soon.
        if success:
            return "Sent command to pdinstall"
        else:
            return "Sending command to pdinstall failed"


    @classmethod
    def set_update_fetcher(cls, update_fetcher):
        WampSession.update_fetcher = update_fetcher


    @inlineCallbacks
    def updatesPending(self, pdid):
        out.info('Notified of updates...')
        if WampSession.update_fetcher is not None:
            out.info('Pulling updates from the server...')
            yield WampSession.update_fetcher.pull_update()



from ruamel.yaml import YAML

from paradrop.core.auth.user import User
from paradrop.core.chute.chute import Chute
from paradrop.core.chute.service import Service


yaml = YAML(typ='safe')

yaml.default_flow_style = False

yaml.register_class(User)
yaml.register_class(Chute)
yaml.register_class(Service)



from twisted.internet.protocol import ProcessProtocol
from paradrop.base.output import out

class AnalyzerProcessProtocol(ProcessProtocol):
    def __init__(self, airshark_manager):
        self.ready = False
        self.airshark_manager = airshark_manager

    def isRunning(self):
        return self.ready

    def connectionMade(self):
        out.info('Airshark analyzer process starts')
        self.ready = True

    def childDataReceived(self, childFd, data):
        if (childFd == 4):
            # Output of airshark analyzer
            #out.info(data)
            self.airshark_manager.on_analyzer_message(data)
        elif (childFd == 1 or childFd == 2):
            # stdout/stderr of airshark analyzer
            out.info('Airshark: ========\n%s==================' % data)

    def processEnded(self, status):
        out.info('Airshark analyzer process exits')
        self.ready = False

    def feedSpectrumData(self, data):
        self.transport.writeToChild(3, data)

    def stop(self):
        self.transport.signalProcess('KILL')



"""
This module downloads a package from a given URL using one of potentially
many different methods.  We currently support the github web API and
simple HTTP(S).  The github method is more developed and returns meta data
about the project (the commit hash and message), but support for other
methods, e.g. download a tar file that was uploaded to a web server,
are not precluded.

Private downloads are supported with the HTTP Authorization header.
For github, we need to use the github API to request a token to access
the owner's private repository.  That part is not implemented here.
"""

import base64
import json
import os
import re
import shutil
import subprocess
import tarfile
import tempfile

import pycurl
import six


from paradrop.base import settings


github_re = re.compile("^(http|https)://github.com/([\w\-]+)/([\w\-\.]+?)(\.git)?$")
general_url_re = re.compile("(http:\/\/|https:\/\/)(\S+)")
hash_re = re.compile("^.*-([0-9a-f]+)$")


class Downloader(object):
    def __init__(self, url, user=None, secret=None, repo_owner=None, repo_name=None):
        self.url = url
        self.user = user
        self.secret = secret
        self.repo_owner = repo_owner
        self.repo_name = repo_name
        self.checkout = None

        self.commitHash = None
        self.commitMessage = None

        self.workDir = None

    def __enter__(self):
        self.workDir = tempfile.mkdtemp()
        return self

    def __exit__(self, type, value, traceback):
        if self.workDir is not None:
            shutil.rmtree(self.workDir)
            self.workDir = None

    def download(self):
        raise NotImplementedError

    def meta(self):
        raise NotImplementedError

    def fetch(self):
        """
        Download the project.

        Returns the full path to the temporary directory containing the project
        and a dictionary containing meta data.
        """
        if self.workDir is None:
            self.workDir = tempfile.mkdtemp()
        runDir = self.download()
        meta = self.meta()
        return runDir, meta

    def extract(self):
        tar = tarfile.open(self.tarFile)

        # Look for a Dockerfile and also check for dangerous paths (.. or /).
        runPath = None
        for member in tar:
            path = os.path.normpath(member.name)
            if path.startswith(".."):
                raise Exception("Archive contains a forbidden path: {}".format(path))
            elif os.path.isabs(path):
                raise Exception("Archive contains an absolute path: {}".format(path))
            elif path.endswith(settings.CHUTE_CONFIG_FILE):
                runPath = path
            elif path.endswith("Dockerfile"):
                runPath = path
            elif self.commitHash is None:
                match = hash_re.match(path)
                if match is not None:
                    self.commitHash = match.group(1)

        tar.extractall(path=self.workDir)

        if runPath is None:
            raise Exception("Repository does not contain {} or Dockerfile".format(
                settings.CHUTE_CONFIG_FILE))

        relRunDir = os.path.dirname(runPath)
        runDir = os.path.join(self.workDir, relRunDir)
        return runDir


class GitSSHDownloader(Downloader):
    def __init__(self, url, checkout="master", **kwargs):
        """
        checkout: branch, tag, or commit hash to checkout (default: "master").
        """
        super(GitSSHDownloader, self).__init__(url, **kwargs)

        if checkout:
            self.checkout = checkout
        else:
            # Interpret None or empty string as the default, "master".
            self.checkout = "master"

    def download(self):
        env = os.environ.copy()
        key_file = os.path.join(settings.KEY_DIR, "node.key")

        if os.path.isfile(key_file):
            # TODO: Set up a way for the node to securely retrieve the host key
            # from the server and add it to the known hosts file. This would
            # probably need to go through the web server, which uses HTTPS.
            env['GIT_SSH_COMMAND'] = "ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i {}".format(key_file)

        cmd = ["git", "clone", self.url, self.workDir]
        subprocess.call(cmd, env=env)

        cmd = ["git", "-C", self.workDir, "checkout", self.checkout]
        subprocess.call(cmd)

        return self.workDir

    def meta(self):
        return {}


class GithubDownloader(Downloader):
    def __init__(self, url, checkout="master", **kwargs):
        """
        checkout: branch, tag, or commit hash to checkout (default: "master").
        """
        super(GithubDownloader, self).__init__(url, **kwargs)

        if checkout:
            self.checkout = checkout
        else:
            # Interpret None or empty string as the default, "master".
            self.checkout = "master"

    def _create_curl_conn(self, url):
        """
        Create a cURL connection object with useful default settings.
        """
        headers = []
        if self.user is not None and self.secret is not None:
            b64cred = base64.b64encode("{}:{}".format(self.user, self.secret))
            headers.append("Authorization: Basic {}".format(b64cred))

        conn = pycurl.Curl()

        if len(headers) > 0:
            conn.setopt(pycurl.HTTPHEADER, headers)

        conn.setopt(pycurl.URL, url)

        # github often redirects
        conn.setopt(pycurl.FOLLOWLOCATION, 1)

        return conn

    def download(self):
        url = "https://github.com/{}/{}/tarball/{}".format(
                self.repo_owner, self.repo_name, self.checkout)
        conn = self._create_curl_conn(url)

        self.tarFile = os.path.join(self.workDir, "source.tar.gz")
        with open(self.tarFile, "w") as output:
            conn.setopt(pycurl.WRITEFUNCTION, output.write)
            conn.perform()

        http_code = conn.getinfo(pycurl.HTTP_CODE)
        if http_code != 200:
            raise Exception("Error downloading archive: response {}".format(http_code))

        return self.extract()

    def meta(self):
        """
        Return repository meta data as a dictionary.
        """
        result = {}

        if self.commitHash is not None:
            result['CommitHash'] = self.commitHash
        if self.commitMessage is not None:
            result['CommitMessage'] = self.commitMessage

        # If set, self.commitHash may be more specific than self.checkout (e.g.
        # commit hash vs. branch name).  It is better to use the most specific
        # one to query for meta data.
        checkout = self.commitHash
        if checkout is None:
            checkout = self.checkout

        url = "https://api.github.com/repos/{owner}/{repo}/commits/{sha}".format(
                owner=self.repo_owner, repo=self.repo_name, sha=checkout)
        conn = self._create_curl_conn(url)

        response = six.StringIO()
        conn.setopt(pycurl.WRITEFUNCTION, response.write)
        conn.perform()

        http_code = conn.getinfo(pycurl.HTTP_CODE)
        if http_code == 200:
            data = json.loads(response.getvalue())
            result['Commit'] = data['commit']
            result['CommitMessage'] = data['commit']['message']

        return result


class WebDownloader(Downloader):
    def _create_curl_conn(self, url):
        """
        Create a cURL connection object with useful default settings.
        """
        headers = []
        if self.user is not None and self.secret is not None:
            b64cred = base64.b64encode("{}:{}".format(self.user, self.secret))
            headers.append("Authorization: Basic {}".format(b64cred))

        conn = pycurl.Curl()

        if len(headers) > 0:
            conn.setopt(pycurl.HTTPHEADER, headers)

        conn.setopt(pycurl.URL, url)

        # github often redirects
        conn.setopt(pycurl.FOLLOWLOCATION, 1)

        return conn

    def download(self):
        conn = self._create_curl_conn(self.url)

        self.tarFile = os.path.join(self.workDir, "source.tar")
        with open(self.tarFile, "w") as output:
            conn.setopt(pycurl.WRITEFUNCTION, output.write)
            conn.perform()

        http_code = conn.getinfo(pycurl.HTTP_CODE)
        if http_code != 200:
            raise Exception("Error downloading archive: response {}".format(http_code))

        return self.extract()

    def meta(self):
        """
        Return repository meta data as a dictionary.
        """
        result = {}

        return result


def downloader(url, user=None, secret=None, **kwargs):
    """
    Return an appropriate Downloader for the given URL.

    This should be used in a "with ... as ..." statement to perform cleanup on
    all exit cases.

    Example:
    with downloader("https://github.com/...") as dl:
        path, meta = dl.fetch()
        # do some work on the repo here
    """
    # If the URL looks like a github.com reop, then use github-specific
    # download method.  Otherwise, use basic web download method.
    match = github_re.match(url)
    if match is not None:
        if user is None:
            user = match.group(2)
        repo_owner = match.group(2)
        repo_name = match.group(3)
        return GithubDownloader(url, user=user, secret=secret,
                repo_owner=repo_owner, repo_name=repo_name, **kwargs)

    # If the URL starts with ssh://, then use the git SSH download method.
    if url.startswith("ssh://"):
        return GitSSHDownloader(url, **kwargs)

    return WebDownloader(url, user=user, secret=secret, **kwargs)



from __future__ import absolute_import
from twisted.internet.task import LoopingCall
from twisted.internet.reactor import spawnProcess

from paradrop.base.output import out
from paradrop.base import settings
from paradrop.lib.utils import pdos
from paradrop.core.config.airshark import airshark_interface_manager
from .scanner import Scanner
from .analyzer import AnalyzerProcessProtocol


class AirsharkManager(object):
    def __init__(self):
        self.loop = LoopingCall(self.check_spectrum)
        self.wifi_interface = None
        self.scanner = None
        self.analyzer_process = AnalyzerProcessProtocol(self)
        self.spectrum_observers = []
        self.analyzer_observers = []

        airshark_interface_manager.add_observer(self)

    def status(self):
        hardware_ready = False
        software_ready = False
        airshark_running = False
        if self.wifi_interface:
            hardware_ready = True

        if pdos.exists(settings.AIRSHARK_INSTALL_DIR):
            software_ready = True

        if self.analyzer_process.isRunning():
            airshark_running = True

        return (hardware_ready, software_ready, airshark_running)

    def on_interface_up(self, interface):
        self._stop()
        self.wifi_interface = interface
        self.scanner = Scanner(self.wifi_interface)
        self._start()

    def on_interface_down(self, interface):
        self._stop()
        self.wifi_interface = None
        self.scanner = None

    def _start(self):
        self.scanner.cmd_chanscan()
        self.scanner.start()

        if not self.analyzer_process.isRunning():
            out.info("Launching airshark analyzer")
            cmd = [settings.AIRSHARK_INSTALL_DIR + "/analyzer", settings.AIRSHARK_INSTALL_DIR + "/specshape/specshape_v1_722N_5m.txt", "--spectrum-fd=3", "--output-fd=4"]
            spawnProcess(self.analyzer_process,\
                         cmd[0], cmd, env=None, \
                         childFDs={0:"w", 1:"r", 2:2, 3:"w", 4:"r"})

        self.loop.start(0.2)
        return True

    def _stop(self):
        if self.scanner:
            self.scanner.stop()

        if self.analyzer_process.isRunning():
            self.analyzer_process.stop()

        if self.loop.running:
            self.loop.stop()

    def check_spectrum(self):
        # The bandwidth of the data is about 160k Bytes per second
        ts, data = self.scanner.spectrum_reader.read_samples()
        if ((data is not None) and (len(self.spectrum_observers) > 0 or self.analyzer_process.isRunning())):
            if len(self.spectrum_observers) > 0:
                #for (tsf, max_exp, freq, rssi, noise, max_mag, max_index, bitmap_weight, sdata) in SpectrumReader.decode(data):
                #    for observer in self.spectrum_observers:
                #        observer.on_spectrum_data(tsf, max_exp, freq, rssi, noise, max_mag, max_index, bitmap_weight, sdata)

                # Due to performance issue, we have to delegate the packet decoding task to clients
                # for packet in SpectrumReader.decode(data):
                #    for observer in self.spectrum_observers:
                #        observer.on_spectrum_data(packet)
                for observer in self.spectrum_observers:
                    observer.on_spectrum_data(data)

            if self.analyzer_process.isRunning():
                # Forward spectrum data to the airshark analyzer
                self.analyzer_process.feedSpectrumData(data)

    # TODO: Not sure we need it or not
    def read_raw_samples(self):
        if (self.scanner):
            return self.scanner.spectrum_reader.read_samples()
        else:
            return None, None

    def on_analyzer_message(self, message):
        for observer in self.analyzer_observers:
            observer.on_analyzer_message(message)

    def add_spectrum_observer(self, observer):
        if (self.spectrum_observers.count(observer) == 0):
            self.spectrum_observers.append(observer)

    def remove_spectrum_observer(self, observer):
        if (self.spectrum_observers.count(observer) == 1):
            self.spectrum_observers.remove(observer)

    def add_analyzer_observer(self, observer):
        if (self.analyzer_observers.count(observer) == 0):
            self.analyzer_observers.append(observer)

    def remove_analyzer_observer(self, observer):
        if (self.analyzer_observers.count(observer) == 1):
            self.analyzer_observers.remove(observer)



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################
'''
    This module contains the methods required to generate and act upon
    execution plans.

    An execution plan is a set of operations that must be performed
    to update a Chute from some old state into the new state provided
    by the API server.

    All plans that are generated are function pointers, as in no actual
    operations are performed during the generation process.
'''

import traceback

from twisted.internet.defer import Deferred

from paradrop.base.output import out


def generatePlans(update):
    """
    For an update object provided this function references the updateModuleList which lets all exc
    modules determine if they need to add functions to change the state of the system when new 
    chutes are added to the OS.

    Returns: True in error, as in we should stop with this update plan
    """
    out.header('Generating %r\n' % (update))

    # Iterate through the list provided for this update type
    for mod in update.updateModuleList:
        if(mod.generatePlans(update)):
            return True


def aggregatePlans(update):
    """
        Takes the PlanMap provided which can be a combination of changes for multiple
        different chutes and it puts things into a sane order and removes duplicates where
        possible.

        This keeps things like reloading networking from happening twice if 2 chutes make changes.

        Returns:
            A new PlanMap that should be executed
    """
    out.header('Aggregating plans\n')
    # For now we just order the plans and return a new list
    update.plans.sort()


def executePlans(update):
    """
        Primary function that actually executes all the functions that were added to plans by all
        the exc modules. This function can heavily modify the OS/files/etc.. so the return value is
        very important.
        Returns:
            True in error : abortPlans function should be called
            False otherwise : everything is OK
    """
    out.header('Executing plans %r\n' % (update))
    # Finding the functions to call is actually done by a 'iterator' like function in the plangraph module
    while(True):
        # This function either returns None or a tuple just like generate added to it
        p = update.plans.getNextTodo()

        # No more to do?
        if(not p):
            break

        # Explode tuple otherwise
        func, args = p

        # We are in a try-except block so if func isn't callable that will catch it
        try:
            out.verbose('Calling %s\n' % (func))
            update.progress("Calling {}".format(func.__name__))
            #
            # Call the function from the execution plan
            #
            # args may be empty, but we don't want to pass in a tuple if we don't need to.
            # This below explodes the args so if @args is (), then what is passed is @update
            skipme = func(*((update, ) + args))

        except Exception as e:
            out.exception(e, True)
            # plans = str(update.plans)) # Removed because breaks new out.exception call
            out.warn("Failed to execute plan %s%s" % (func.__name__, args))
            update.responses.append({'exception': str(e), 'traceback': traceback.format_exc()})
            update.failure = str(e)
            return True

        # The functions we call here can return other functions, if they do
        # these are functions that should be skipped later on (for instance a
        # set* function discovering it didn't change anything, later on we
        # shouldn't call the corresponding reload function)
        if skipme is not None:
            # If the function returned a Deferred, we will drop out of the
            # execution pipeline and resume later.
            if isinstance(skipme, Deferred):
                out.verbose('Function {} returned a Deferred'.format(func))
                return skipme

            # These functions can return individual functions to skip, or a
            # list of multiple functions
            elif callable(skipme):
                skipme = [skipme]

            for skip in skipme:
                out.warn('Identified a skipped function: %r\n' % (skip))
                update.plans.registerSkip(skip)

    # Now we are done
    return False


def abortPlans(update):
    """
        This function should be called if one of the Plan objects throws an Exception.
        It takes the PlanMap argument and calls the getNextAbort function just like executePlans does with todoPlans.
        This dynamically generates an abort plan list based on what plans were originally executed.
        Returns:
            True in error : This is really bad
            False otherwise : we were able to restore system state back to before the executeplans function was called
    """
    out.header('Aborting plans %r\n' % (update.plans))
    sameError = False
    while(True):
        # This function either returns None or a tuple just like generate added to it
        p = update.plans.getNextAbort()

        # No more to do?
        if(not p):
            break

        # Explode tuple otherwise
        func, args = p

        # We are in a try-except block so if func isn't callable that will catch it
        try:
            out.verbose('Calling {}\n'.format(func))
            update.progress('Calling {}'.format(func.__name__))

            func(*((update, ) + args))

            # If the func is called without exception then clear the @sameError flag for the next function call
            sameError = False

        except Exception as e:
            # Since we are running this in an infinite loop if a major function throws an error
            # we could loop forever, so check for the error, which is only reset at the end of the loop
            if(sameError):
                return True
            update.responses.append({'exception': str(e), 'traceback': traceback.format_exc()})
            out.fatal('An abort function raised an exception!!! %r: %s\n%s\n' % (update.plans, str(e), traceback.format_exc()))
            sameError = True

    # Getting here we assume the system state has been restored using our abort plan
    return False



###################################################################
# Copyright 2013-2016 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
This module generates update plans for a host configuration operation.  It is
separate from the modules that generate plans for chute operations because we
only need to do a subset of the operations.
"""

from paradrop.base.output import out
from paradrop.core.config import airshark, devices, haproxy, network, configservice, hostconfig, reservations, services, zerotier

from . import plangraph


# Update types that apply changes from host configuration.
CONFIG_CHANGING_TYPES = set([
    "inithostconfig",
    "sethostconfig",
    "patchhostconfig"
])


def generatePlans(update):
    out.verbose("%r\n" % (update))

    # Detect system devices and set up basic configuration for them (WAN
    # interface, wireless devices).  These steps do not need to be reverted on
    # abort.
    #
    # checkSystemDevices may reboot the machine if expected devices are
    # missing.
    #
    # abortNetworkConfig is added as an abort command here so that it runs when
    # config.network.getNetworkConfig or just about anything else fails.
    #
    # reloadAll is added as an abort command here so that it runs when any of
    # the set* plans fail and back out.
    update.plans.addPlans(plangraph.STRUCT_GET_SYSTEM_DEVICES,
                          (devices.getSystemDevices, ),
                          (network.abortNetworkConfig, ))
    update.plans.addPlans(plangraph.CHECK_SYSTEM_DEVICES,
                          (devices.checkSystemDevices, ))

    update.plans.addPlans(plangraph.STRUCT_GET_RESERVATIONS,
                          (reservations.getReservations, ))

    update.plans.addPlans(plangraph.STRUCT_GET_HOST_CONFIG,
                          (hostconfig.getHostConfig, ))

    # Save current network configuration into chute cache (key: 'networkInterfaces')
    update.plans.addPlans(plangraph.STRUCT_GET_INT_NETWORK,
                          (network.getNetworkConfig, ))

    # Start haproxy.  This does not depend on the host config, and we want
    # it setup even in cases where applying the host config failed.
    update.plans.addPlans(plangraph.RECONFIGURE_PROXY,
                          (haproxy.reconfigureProxy, ))

    if update.updateType in CONFIG_CHANGING_TYPES:
        # Save current host configuration to disk.
        update.plans.addPlans(plangraph.STRUCT_SET_HOST_CONFIG,
                              (hostconfig.setHostConfig, ),
                              (hostconfig.revertHostConfig, ))

        # Apply host configuration to system configuration.
        update.plans.addPlans(plangraph.STRUCT_SET_SYSTEM_DEVICES,
                              (devices.setSystemDevices, ),
                              (configservice.reloadAll, ))

        # Apply zerotier configuration.
        update.plans.addPlans(plangraph.ZEROTIER_CONFIGURE,
                              (zerotier.configure, ))

        # Apply Airshark configuration.
        update.plans.addPlans(plangraph.AIRSHARK_CONFIGURE,
                              (airshark.configure, ))

        # Configure telemetry service.
        update.plans.addPlans(plangraph.TELEMETRY_SERVICE,
                              (services.configure_telemetry, ))

        # Reload configuration files
        todoPlan = (configservice.reloadAll, )
        update.plans.addPlans(plangraph.RUNTIME_RELOAD_CONFIG, todoPlan)

        # Reload configuration files if aborting.  This needs to happen at the
        # right place in the update pipeline such that UCI files have been
        # restored to their previous contents.
        todoPlan = (configservice.reload_placeholder, )
        abtPlan = (configservice.reloadAll, )
        update.plans.addPlans(plangraph.RUNTIME_RELOAD_CONFIG_BACKOUT, todoPlan, abtPlan)



def str_to_numeric(s):
    """
    Convert a string to a numeric type.

    Returns either an int or a float depending on the apparent type of the
    string.
    """
    try:
        return int(s)
    except ValueError:
        return float(s)



"""
The ProcessMonitor class ensures that a service is running and that its pid
file is consistent.

This addresses an issue we have had with Docker on Ubuntu Snappy, where its
pid file sometimes persists and prevents the service from starting.
"""
from __future__ import print_function

import glob
import os
import subprocess
import time

import psutil


class ProcessMonitor(object):
    # Specify allowed corrective actions, which we can change when running
    # locally to disable rebooting, for example.
    #
    # TODO: Implement a more general module for checking system health and
    # applying corrective action.
    allowedActions = set(["restart", "reboot"])

    def __init__(self, service, cmdstring=None, pidfile=None, action="restart"):
        """
        service: service name (used to restart it).
        cmdstring: string to look for in running command name (e.g. "docker")
        pidfile: None or path to look for pid file(s).
        Bash-style globbing is supported, e.g. "/var/snap/docker/*/run/docker.pid".
        action: "restart" the service or "reboot" the machine
        """
        self.service = service
        self.action = action

        if cmdstring is not None:
            self.cmdstring = cmdstring
        else:
            self.cmdstring = service

        if pidfile is not None:
            self.pidfiles = [ pidfile ]
        else:
            self.pidfiles = [
                "/var/snap/{service}/current/run/{service}.pid".format(service=service)]

    def check(self):
        """
        Check that the service is running and consistent with pid file(s).

        Returns True or False.
        """
        # Set of pids (strings) where the command string matches what we are
        # looking for.
        detected_pids = set()

        # Set of pids (strings) that are both running processes and found in
        # pid files.
        consistent_pids = set()

        # Search for running processes that match our command string.
        for proc in psutil.process_iter():
            try:
                if self.cmdstring in proc.name():
                    detected_pids.add(str(proc.pid))

            # We could also get psutil.ZombieProcess or
            # psutil.AccessDenied.  We want those to be logged.
            except psutil.NoSuchProcess:
                pass

        # Search for pid file(s) and check consistency.
        for pidfile in self.pidfiles:
            for path in glob.iglob(pidfile):
                with open(path, 'r') as source:
                    pid = source.read().strip()

                if pid in detected_pids:
                    consistent_pids.add(pid)
                else:
                    # Delete the stale pid file.
                    os.remove(path)

        return len(consistent_pids) > 0

    def restart(self):
        """
        Restart the service.
        """
        if self.action == "restart":
            cmd = ["snappy", "service", self.service, "restart"]
        else:
            cmd = ["shutdown", "-r", "now"]

        if self.action in ProcessMonitor.allowedActions:
            print("Running \"{}\" to fix {}".format(" ".join(cmd), self.service))
            return subprocess.call(cmd)
        else:
            print("Warning: would run \"{}\" to fix {}, but not allowed.".format(
                " ".join(cmd), self.service))

    def ensureReady(self, delay=5, tries=3):
        """
        Look through checking and restarting the service until it is ready or
        the maximum number of tries has been reached.

        delay: time delay (seconds) between retries.
        tries: maximum number of restart-wait-check cycles.
        """
        ready = self.check()
        if ready:
            return True

        for t in range(tries):
            time.sleep(delay)
            ready = self.check()
            if ready:
                return True
            else:
                self.restart()

        time.sleep(delay)
        return self.check()


dockerMonitor = ProcessMonitor("docker", action="reboot")
containerdMonitor = ProcessMonitor("docker-containerd",
        pidfile="/var/snap/docker/current/run/docker/libcontainerd/docker-containerd.pid",
        action="reboot")



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.config import resource
from paradrop.core.container import dockerapi
from . import plangraph

def generatePlans(update):
    """
        This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
        then adds Plan() calls to make the Chute match the @newChute.

        Returns:
            True: abort the plan generation process
    """
    out.verbose("%r\n" % (update))

    update.plans.addPlans(plangraph.RESOURCE_GET_ALLOCATION,
            (resource.getResourceAllocation,))

    update.plans.addPlans(plangraph.RESOURCE_SET_ALLOCATION,
            (dockerapi.setResourceAllocation,),
            (dockerapi.revertResourceAllocation,))

    return None



###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import time
import threading
from twisted.internet import defer, threads

from paradrop.base.output import out
from paradrop.base.pdutils import timeint
from paradrop.base import constants, nexus, settings
from paradrop.core.agent import reporting
from paradrop.lib.misc.procmon import dockerMonitor, containerdMonitor
from paradrop.core.chute.restart import reloadChutes

from . import update_object


class UpdateManager:

    """
        This class is in charge of making the configuration changes required on the chutes.
        It utilizes the ChuteStorage class to hold onto the chute data.

        Use @updateChutes to make the configuration changes on the AP.
            This function is thread-safe, this class will only call one update set at a time.
            All others are held in a queue until the last update is complete.
    """

    def __init__(self, reactor):
        self.reactor = reactor

        self.updateLock = threading.Lock()
        self.updateQueue = []

        # Map update_id -> update object.
        self.active_changes = {}

        # TODO: Ideally, load this from file so that change IDs are unique
        # across system reboots.
        self.next_change_id = 1

        ###########################################################################################
        # Launch the first update call, NOTE that you have to use callInThread!!
        # This happens because the perform_updates should run in its own thread,
        # it makes blocking calls and such... so if we *don't* use callInThread
        # then this function WILL BLOCK THE MAIN EVENT LOOP (ie. you cannot send any data)
        #
        # FIXME: It should be noted that calling updateLock.acquire(), e.g.
        # from add_update, also blocks the main thread.  Synchronizing access
        # to the work queue should be reworked.
        ###########################################################################################
        self.reactor.callInThread(self._perform_updates)

    def _get_next_update(self):
        """MUTEX: updateLock
            Returns the size of the local update queue.
        """
        self.updateLock.acquire()
        if(len(self.updateQueue) > 0):
            # Get first available
            a = self.updateQueue.pop(0)
        else:
            a = None
        self.updateLock.release()
        return a

    def clear_update_list(self):
        """MUTEX: updateLock
            Clears all updates from list (new array).
        """
        self.updateLock.acquire()
        self.updateQueue = []
        self.updateLock.release()

    def add_update(self, **update):
        """MUTEX: updateLock
            Take the list of Chutes and push the list into a queue object, this object will then call
            the real update function in another thread so the function that called us is not blocked.

            We take a callable responseFunction to call, when we are done with this update we should call it.
        """
        d = defer.Deferred()
        update['deferred'] = d

        # Make sure the update object has a change ID for message retrieval.  Some
        # code paths (API code) will prefer to set the change ID before calling
        # add_update so that they can return the change ID to the user.
        if 'change_id' not in update:
            update['change_id'] = self.assign_change_id()

        # Convert to Update object before storing.
        updateObj = update_object.parse(update)

        self.updateLock.acquire()
        self.updateQueue.append(updateObj)
        self.updateLock.release()

        return d

    def add_provision_update(self, hostconfig_patch, zerotier_networks):
        update = self._make_router_update("patchhostconfig")
        update.patch = list(hostconfig_patch)

        for network in zerotier_networks:
            update.patch.append({
                "op": "add",
                "path": "/zerotier/networks/-",
                "value": network
            })

        self.updateQueue.append(update)

    def assign_change_id(self):
        """
        Get a unique change ID for an update.

        This should be used to set the change_id field in an update object.
        """
        x = self.next_change_id
        self.next_change_id += 1
        return x

    def find_change(self, change_id):
        """
        Search active and queued changes for the requested change.

        Returns an Update object or None.
        """
        if change_id in self.active_changes:
            return self.active_changes[change_id]

        for update in self.updateQueue:
            if update.change_id == change_id:
                return update

        return None

    def _make_router_update(self, updateType):
        """
        Make a ROUTER class update object.
        """
        update = dict(updateClass='ROUTER',
                    updateType=updateType,
                    name=constants.RESERVED_CHUTE_NAME,
                    tok=timeint())
        return update_object.parse(update)

    def _perform_updates(self):
        """This is the main working function of the PDConfigurer class.
            It should be executed as a separate thread, it does the following:
                checks for any updates to perform
                does them
                responds to the server
                removes the update
                checks for more updates
                    if more exist it calls itself again more quickly
                    else it puts itself to sleep for a little while
        """
        if settings.CHECK_DOCKER:
            ready = dockerMonitor.ensureReady()
            if not ready:
                out.warn("Docker does not appear to be running.  "
                            "Most functionality with containers will be broken.")

            ready = containerdMonitor.ensureReady()
            if not ready:
                out.warn("Docker containerd does not appear to be running.  "
                            "Most functionality with containers will be broken.")

        # add any chutes that should already be running to the front of the
        # update queue before processing any updates
        startQueue = reloadChutes()
        self.updateLock.acquire()
        self.updateQueue.append(self._make_router_update("prehostconfig"))
        self.updateQueue.append(self._make_router_update("inithostconfig"))
        self.updateQueue.extend(startQueue)
        self.updateLock.release()

        # Always perform this work
        while self.reactor.running:
            # Check for new updates
            change = self._get_next_update()
            if change is None:
                time.sleep(1)
                continue

            self._perform_update(change)

            # Apply a batch of updates and when the queue is empty, send a
            # state report.  We're not reacquiring the mutex here because the
            # worst case is we send out an extra state update.
            if len(self.active_changes) == 0 and nexus.core.provisioned():
                threads.blockingCallFromThread(self.reactor,
                        reporting.sendStateReport)

    def _perform_update(self, update):
        """
        Perform a single update, to be called by perform_updates.

        This is split from perform_updates for easier unit testing.
        """
        # Add to the active set when processing starts. It is a dictionary, so
        # it does not matter if this is the first time we see this update or
        # if we are resuming it.
        self.active_changes[update.change_id] = update

        try:
            # Mark update as having been started.
            update.started()
            out.info('Performing update %s\n' % (update))

            # TESTING start
            # TODO: still need this?
            if(settings.FC_BOUNCE_UPDATE): # pragma: no cover
                out.testing('Bouncing update %s, result: %s\n' % (
                    update, settings.FC_BOUNCE_UPDATE))
                update.complete(success=True, message=settings.FC_BOUNCE_UPDATE)
                return
            # TESTING end

            # Based on each update type execute could be different
            result = update.execute()
            if isinstance(result, defer.Deferred):
                # Update is not done, but it is yielding. When the deferred
                # fires, add it back to the work queue to resume it.
                #
                # TODO: We could handle errors separately and use that to break
                # the update pipeline, but then we need to propagate that
                # through the update.execute, executionplan chain.  For now,
                # we have an update stage right after prepare_image that checks
                # if the build was successful or throws an exception. That
                # should work but is not very general.
                def resume(result):
                    self.updateQueue.append(update)
                result.addBoth(resume)
            elif update.change_id in self.active_changes:
                # Update is done, so remove it from the active list.
                del self.active_changes[update.change_id]

        except Exception as e:
            out.exception(e, True)



###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import errno
import os
import subprocess
import shutil

import six

from distutils import dir_util

# We have to import this for the decorator
from paradrop.base.output import out


# protect the original open function
__open = open

# Since we overwrite everything else, do the same to basename
basename = lambda x: os.path.basename(x)

def getMountCmd():
    return "mount"


def isMount(mnt):
    """This function checks if @mnt is actually mounted."""
    # TODO - need to check if partition and mount match the expected??
    return os.path.ismount(mnt)


def oscall(cmd, get=False):
    """
    This function performs a OS subprocess call.
    All output is thrown away unless an error has occured or if @get is True
    Arguments:
        @cmd: the string command to run
        [get] : True means return (stdout, stderr)
    Returns:
        None if not @get and no error
        (stdout, retcode, stderr) if @get or yes error
    """
    # Since we are already in a deferred chain, use subprocess to block and make the call to mount right HERE AND NOW
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, errors = proc.communicate()
    if(proc.returncode or get):
        return (output, proc.returncode, errors)
    else:
        if(output and output != ""):
            out.verbose('"%s" stdout: "%s"\n' % (cmd, output.rstrip()))
        if(errors and errors != ""):
            out.verbose('"%s" stderr: "%s"\n' % (cmd, errors.rstrip()))
        return None


def getFileType(f):
    if not exists(f):
        return None
    r = oscall('file "%s"' % f, True)
    if(r is not None and isinstance(r, tuple)):
        return r[0]
    else: # pragma: no cover
        return None


def exists(p):
    return os.path.exists(p)


def listdir(p):
    return os.listdir(p)


def unlink(p):
    return os.unlink(p)


def mkdir(p):
    return os.mkdir(p)


def symlink(a, b):
    return os.symlink(a, b)


def ismount(p):
    return os.path.ismount(p)


def fixpath(p):
    """This function is required because if we need to pass a path to something like tarfile,
        we cannot overwrite the function to fix the path, so we need to expose it somehow."""
    return p


def copy(a, b):
    return shutil.copy(a, b)


def move(a, b):
    return shutil.move(a, b)


def remove(path, suppressNotFound=False):
    if (isdir(path)):
        return shutil.rmtree(path)
    else:
        try:
            os.remove(path)
        except OSError as err:
            # Suppress the exception if it is a file not found error and the
            # suppressNotFound flag is set.  Otherwise, re-raise the exception.
            if not suppressNotFound or err.errno != errno.ENOENT:
                raise


def isdir(a):
    return os.path.isdir(a)


def isfile(a):
    return os.path.isfile(a)


def copytree(a, b):
    """shutil's copytree is dumb so use distutils."""
    return dir_util.copy_tree(a, b)


def open(p, mode):
    return __open(p, mode)


def writeFile(filename, line, mode="a"):
    """Adds the following cfg (either str or list(str)) to this Chute's current
        config file (just stored locally, not written to file."""
    try:
        if isinstance(line, list):
            data = "\n".join(line) + "\n"
        elif isinstance(line, six.string_types):
            data = "%s\n" % line
        else:
            out.err("Bad line provided for %s\n" % filename)
            return
        fd = open(filename, mode)
        fd.write(data)
        fd.flush()
        fd.close()

    except Exception as e:
        out.err('Unable to write file: %s\n' % (str(e)))


def write(filename, data, mode="w"):
    """ Writes out a config file to the specified location.
    """
    try:
        fd = open(filename, mode)
        fd.write(data)
        fd.flush()
        fd.close()
    except Exception as e:
        out.err('Unable to write to file: %s\n' % str(e))


def readFile(filename, array=True, delimiter="\n"):
    """
        Reads in a file, the contents is NOT expected to be binary.
        Arguments:
            @filename: absolute path to file
            @array : optional: return as array if true, return as string if False
            @delimiter: optional: if returning as a string, this str specifies what to use to join the lines

        Returns:
            A list of strings, separated by newlines
            None: if the file doesn't exist
    """
    if(not exists(filename)):
        return None

    lines = []
    with open(filename, 'r') as fd:
        while(True):
            line = fd.readline()
            if(not line):
                break
            lines.append(line.rstrip())
    if(array is True):
        return lines
    else:
        return delimiter.join(lines)


def read_sys_file(path, default=None):
    """
    Read a file and return the contents as a string.

    This is best suited for files that store a single line of text such as
    files in /sys/.

    Returns the default value if an error occurs.
    """
    try:
        with open(path, 'r') as source:
            return source.read().strip()
    except:
        return default



from .base import ConfigObject, ConfigOption
from .command import Command, KillCommand


class ConfigBridge(ConfigObject):
    typename = "bridge"

    options = [
        ConfigOption(name="interfaces", type=list, required=True)
    ]

    def apply(self, allConfigs):
        commands = list()

        cmd = ["parprouted"]
        for section_name in self.interfaces:
            interface = self.lookup(allConfigs, "network", "interface", section_name)
            cmd.append(interface.config_ifname)

        self.daemon_command = Command(cmd, self)
        commands.append((self.PRIO_START_DAEMON, self.daemon_command))

        return commands

    def revert(self, allConfigs):
        commands = list()

        commands.append((-self.PRIO_START_DAEMON,
            KillCommand(self.daemon_command.pid, self)))

        return commands



'''
Get system information
'''

import os
import pkg_resources


DMI_FIELDS = [
    "bios_date",
    "bios_vendor",
    "bios_version",
    "product_name",
    "product_serial",
    "product_version",
    "sys_vendor"
]


def getOSVersion():
    """
    Return a string identifying the host OS.
    """
    try:
        with open('/proc/version_signature', 'r') as source:
            return source.read().strip()
    except:
        return None


def getPackageVersion(name):
    """
    Get a python package version.

    Returns: a string or None
    """
    try:
        pkg = pkg_resources.get_distribution(name)
        return pkg.version
    except:
        return None


def getDMI():
    """
    Read hardware information from DMI.

    This function attempts to read from known files in /sys/class/dmi/id/.  If
    any are missing or an error occurs, those fields will be omitted from the
    result.

    Returns: a dictionary with fields such as bios_version and product_serial.
    """
    dmi = dict()

    for field in DMI_FIELDS:
        path = os.path.join("/sys/class/dmi/id", field)
        try:
            with open(path, 'r') as source:
                value = source.read().strip()
                dmi[field] = value
        except:
            pass

    return dmi



'''
Fetch new updates from the pdserver and apply the updates
'''
from __future__ import print_function

from twisted.internet import task
from twisted.internet.defer import inlineCallbacks

from paradrop.core.agent.http import PDServerRequest
from paradrop.core.auth.user import User
from paradrop.base.output import out
from paradrop.base.pdutils import timeint


# Interval (in seconds) for polling the server for updates.  This is a fallback
# in case the notifications over WAMP fail, so it can be relatively infrequent.
#
# The automatic polling does not start until after pull_update()
# has been called at least once.
UPDATE_POLL_INTERVAL = 15 * 60


class UpdateFetcher(object):
    def __init__(self, update_manager):
        # Store set of update IDs that are in progress in order to avoid
        # repeats.
        self.updates_in_progress = set()
        self.update_manager = update_manager
        self.scheduled_call = task.LoopingCall(self.pull_update, _auto=True)

        self.long_poll_started = False
        self.use_long_poll = True

    @inlineCallbacks
    def start_long_poll(self):
        self.long_poll_started = True

        while self.use_long_poll:
            request = PDServerRequest('/api/routers/{router_id}/updates/poll')
            try:
                response = yield request.get()
            except Exception:
                pass
            else:
                # 200 = update(s) available
                # 204 = no updates yet
                # 404 = server does not support this endpoint
                if response.code == 200:
                    self.pull_update(_auto=False)
                elif response.code == 404:
                    self.use_long_poll = False

    def start_polling(self):
        # Long polling is a more recent addition. If the server supports it, it
        # will only respond to our GET request after an update is available or a
        # timeout has expired.
        if not self.long_poll_started:
            self.start_long_poll()

        # Basic polling is our baseline. It always works with no dependency on
        # maintaining websocket connections, etc. Set the polling interval to
        # something reasonable, e.g. 15 minutes to 1 hour.
        if not self.scheduled_call.running:
            self.scheduled_call.start(UPDATE_POLL_INTERVAL, now=False)

    @inlineCallbacks
    def pull_update(self, _auto=False):
        """
        Start updates by polling the server for the latest updates.

        This is the only method that needs to be called from outside.  The rest
        are triggered asynchronously.

        Call chain:
        pull_update -> _updates_received -> _update_complete

        _auto: Set to True when called by the scheduled LoopingCall.
        """
        # If this call was triggered externally (_auto=False), then reset the
        # timer for the next scheduled call so we do not poll again too soon.
        if not _auto and self.scheduled_call.running:
            self.scheduled_call.reset()

        def handle_error(error):
            print("Request for updates failed: {}".format(error.getErrorMessage()))

        request = PDServerRequest('/api/routers/{router_id}/updates')
        d = request.get(completed=False)
        d.addCallback(self._updates_received)
        d.addErrback(handle_error)
        yield d


    @inlineCallbacks
    def _updates_received(self, response):
        if not response.success or response.data is None:
            print("There was an error receiving updates from the server.")
            return

        updates = response.data
        print("Received {} update(s) from server.".format(len(updates)))

        # Remove old updates from the set of in progress updates.  These are
        # updates that the server is no longer sending to us.
        received_set = set(item['_id'] for item in updates)
        old_updates = self.updates_in_progress - received_set
        self.updates_in_progress -= old_updates

        for item in updates:
            # TODO: Handle updates with started=True very carefully.
            #
            # There are at least two interesting cases:
            # 1. We started an update but crashed or rebooted without
            # completing it.
            # 2. We started an upgrade to paradrop, the daemon was just
            # restarted, but pdinstall has not reported that the update is
            # complete yet.
            #
            # If we skip updates with started=True, we will not retry updates
            # in case #1.
            if item['_id'] in self.updates_in_progress:
                continue
            elif item.get('started', False):
                yield self._update_ignored(item)
                continue
            else:
                self.updates_in_progress.add(item['_id'])

            creator = item.get('creator', {})
            username = creator.get("name", "paradrop")
            role = creator.get("access_level", "trusted")

            update = dict(updateClass=item['updateClass'],
                          updateType=item['updateType'],
                          user=User(username, "paradrop.org", role=role),
                          tok=timeint())

            # Save fields that relate to external management of this chute.
            update['external'] = {
                'update_id': item['_id'],
                'chute_id': item.get('chute_id', None),
                'version_id': item.get('version_id', None)
            }

            # Backend might send us garbage; be sure that we only accept the
            # config field if it is present and is a dict.
            config = item.get('config', {})
            if isinstance(config, dict):
                update.update(config)

            update = yield self.update_manager.add_update(**update)
            yield self._update_complete(update)


    def _update_complete(self, update):
        """
        Internal: callback after an update operation has been completed
        (successfuly or otherwise) and send a notification to the server.
        """
        # If delegated to an external program, we do not report to pdserver
        # that the update is complete.
        if update.delegated:
            return

        out.info("The update is done, report it to server...")
        update_id = update.external['update_id']
        success = update.result.get('success', False)

        request = PDServerRequest('/api/routers/{router_id}/updates/' +
                str(update_id))
        d = request.patch(
            {'op': 'replace', 'path': '/completed', 'value': True},
            {'op': 'replace', 'path': '/success', 'value': success}
        )

        return d

        # TODO: If this notification fails to go through, we should retry or
        # build in some other mechanism to inform the server.
        #
        # Not catching errors here so we see a stack trace if there is an
        # error.  This is an omission that will need to be dealt with.

    def _update_ignored(self, update):
        """
        Internal: callback for an update that we are ignoring because
        it was started previously and never completed.
        """
        update_id = update['_id']
        request = PDServerRequest('/api/routers/{router_id}/updates/' +
                str(update_id))
        d = request.patch(
            {'op': 'replace', 'path': '/completed', 'value': True},
            {'op': 'replace', 'path': '/success', 'value': False}
        )
        return d



"""
    configservice module:
        This module is responsible for "poking" the proper host OS services
        to change the host OS config. This would include things like changing
        the networking, DHCP server settings, wifi, etc..
"""

import json

from paradrop.confd import client
from paradrop.base.output import out


def reload_placeholder(update):
    """
    This function successfully does nothing.

    It serves as a placeholder so that we can attach an abort function to a
    specific point in the update pipeline.
    """
    pass


def reloadAll(update):
    """
    Reload pdconf configuration files.
    """
    # Note: reloading all config files at once seems safer than individual
    # files because of cross-dependencies.
    statusString = client.reloadAll()

    # Check the status to make sure all configuration sections
    # related to this chute were successfully loaded.
    #
    # We should only abort if there were problems related to this chute.
    # Example: a WiFi card was removed, so we fail to bring up old WiFi
    # interfaces; however, installing a new chute that does not depend on WiFi
    # should still succeed.
    status = json.loads(statusString)
    for section in status:
        # Checking age > 0 filters out errors that occurred in the past.
        if section['success'] or section['age'] > 0:
            continue

        message = "Error installing configuration section {} {} for chute {}".format(
                section['type'], section['name'], section['comment'])
        if section['comment'] == update.new.name:
            raise Exception(message)
        else:
            out.warn(message)



import json
from twisted.internet.defer import succeed
from klein import Klein

from . import cors


class PasswordApi(object):
    '''
    For now, we only support set/reset password for the default user: 'paradrop'
    '''
    routes = Klein()

    def __init__(self, password_manager):
        self.password_manager = password_manager


    @routes.route('/clear', methods=['POST'])
    def clear(self, request):
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        self.password_manager.reset()
        return json.dumps({'success': True})


    # OPTIONS method for CORS
    @routes.route('/change', methods=['POST', 'OPTIONS'])
    def change(self, request):
        cors.config_cors(request)
        if request.method == 'OPTIONS':
            return succeed(None)

        request.setHeader('Content-Type', 'application/json')
        body = json.loads(request.content.read())
        password = body['password']
        result = self.password_manager.change_password(None, password)
        return json.dumps({'success': result})



###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.config import security

from . import plangraph


def generatePlans(update):
    """
    This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
    then adds Plan() calls to make the Chute match the @newChute.

    Returns:
        True: abort the plan generation process
    """
    out.verbose("%r\n" % (update))

    update.plans.addPlans(plangraph.ENFORCE_ACCESS_RIGHTS,
                          (security.enforce_access_rights, ))

    return None



'''
Provides messages from container logs (STDOUT and STDERR).
'''
import os
import signal

import docker
import six

from multiprocessing import Process, Queue


def monitor_logs(service_name, container_name, queue, tail=200):
    """
    Iterate over log messages from a container and add them to the queue
    for consumption.  This function will block and wait for new messages
    from the container.  Use the queue to interface with async code.

    tail: number of lines to retrieve from log history; the string "all"
    is also valid, but highly discouraged for performance reasons.
    """
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')
    container = client.containers.get(container_name)
    output = container.logs(stdout=True, stderr=True,
                            stream=True, timestamps=True, follow=True,
                            tail=tail)
    for line in output:
        # I have grown to distrust Docker streaming functions.  It may
        # return a string; it may return an object.  If it is a string,
        # separate the timestamp portion from the rest of the message.
        if isinstance(line, six.string_types):
            parts = line.split(" ", 1)
            if len(parts) > 1:
                queue.put({
                    'service': service_name,
                    'timestamp': parts[0],
                    'message': parts[1].rstrip()
                })

            else:
                queue.put({
                    'service': service_name,
                    'message': line.rstrip()
                })
        elif isinstance(line, dict):
            line['service'] = service_name
            queue.put(line)

class LogProvider(object):
    def __init__(self, chute):
        self.chute = chute
        self.queue = Queue()
        self.listening = False
        self.processes = []

    def attach(self):
        """
        Start listening for log messages.

        Log messages in the queue will appear like the following:
        {
            'service': 'main',
            'timestamp': '2017-01-30T15:46:23.009397536Z',
            'message': 'Something happened'
        }
        """
        if not self.listening:
            for service in self.chute.get_services():
                process = Process(target=monitor_logs,
                        args=(service.name, service.get_container_name(), self.queue))
                process.start()
                self.processes.append(process)
            self.listening = True

    def get_logs(self):
        logs = []
        while not self.queue.empty():
            msg = self.queue.get()
            logs.append(msg)
        return logs

    def detach(self):
        """
        Stop listening for log messages.

        After this is called, no additional messages will be added to the
        queue.
        """
        if self.listening:
            # We have to kill the process explicitly with SIGKILL,
            # terminate() function does not work here.
            for process in self.processes:
                os.kill(process.pid, signal.SIGKILL)

            self.processes = []
            self.listening = False



"""
"""
from paradrop.lib.misc.governor import GovernorClient


def updateSnap(update):
    client = GovernorClient()
    result = client.updateSnap(update.name, update.data)
    if 'message' in result:
        update.progress(result['message'])



from __future__ import absolute_import
from .json import JSONEncoder


class ParadropJSONEncoder(JSONEncoder):
    def default(self, o):
        if hasattr(o, 'asdict'):
            return o.asdict()

        try:
            return JSONEncoder.default(self, o)
        except TypeError:
            return o.__dict__



from paradrop.core.chute.chute_storage import ChuteStorage


def saveChute(update):
    """
    Save information about the chute to the filesystem.
    """
    chuteStore = ChuteStorage()
    if update.updateType == "delete":
        chuteStore.deleteChute(update.old)
    else:
        chuteStore.saveChute(update.new)


def revertChute(update):
    chuteStore = ChuteStorage()
    if update.updateType == "delete":
        chuteStore.saveChute(update.old)
    elif update.old is not None:
        chuteStore.saveChute(update.old)
    else:
        chuteStore.deleteChute(update.new)


def removeAllChutes(update):
    chuteStore = ChuteStorage()
    chuteStore.clearChuteStorage()



import os
import operator
import random
import re
import string

from .base import ConfigObject, ConfigOption
from .command import Command, KillCommand


CTRL_INTERFACE_DIR = "/tmp/hostapd"
IEEE80211_DIR = "/sys/class/ieee80211"


# Map hardware mode strings from UCI file to hostapd.conf format.
HOSTAPD_HWMODE = {
    '11b': 'b',
    '11g': 'g',
    '11a': 'a'
}


# Map wifi-iface mode values to the type string we pass when running iw
# commands.
IW_IFACE_TYPE = {
    'ap': '__ap',
    'monitor': 'monitor',
    'sta': 'managed'
}


HT40_LOWER_CHANNELS = set([36, 44, 52, 60, 100, 108, 116, 124, 132, 140, 149, 157])
HT40_UPPER_CHANNELS = set([40, 48, 56, 64, 104, 112, 120, 128, 136, 144, 153, 161])

# Map 20 Mhz channel to index of 40 Mhz channel that contains it.
VHT40_CENTER_INDEX = {
    36: 38,
    40: 38,
    44: 46,
    48: 46,
    52: 54,
    56: 54,
    60: 62,
    64: 62,
    100: 102,
    104: 102,
    108: 110,
    112: 110,
    116: 118,
    120: 118,
    124: 126,
    128: 126,
    132: 134,
    136: 134,
    140: 142,
    144: 142,
    149: 151,
    153: 151,
    157: 159,
    161: 159
}

# Map 20 Mhz channel to index of 80 Mhz channel that contains it.
VHT80_CENTER_INDEX = {
    36: 42,
    40: 42,
    44: 42,
    48: 42,
    52: 58,
    56: 58,
    60: 58,
    64: 58,
    100: 106,
    104: 106,
    108: 106,
    112: 106,
    116: 122,
    120: 122,
    124: 122,
    128: 122,
    132: 138,
    136: 138,
    140: 138,
    144: 138,
    149: 155,
    153: 155,
    157: 155,
    161: 155
}

# Map 20 Mhz channel to index of 160 Mhz channel that contains it.
VHT160_CENTER_INDEX = {
    36: 50,
    40: 50,
    44: 50,
    48: 50,
    52: 50,
    56: 50,
    60: 50,
    64: 50,
    100: 114,
    104: 114,
    108: 114,
    112: 114,
    116: 114,
    120: 114,
    124: 114,
    128: 114
}


def getPhyMACAddress(phy):
    path = "{}/{}/macaddress".format(IEEE80211_DIR, phy)
    with open(path, 'r') as source:
        return source.read().strip()


def getPhyFromMAC(mac):
    for phy in os.listdir(IEEE80211_DIR):
        if getPhyMACAddress(phy) == mac:
            return phy
    raise Exception("No wireless device with MAC address {}".format(mac))


def isHexString(data):
    """
    Test if a string contains only hex digits.
    """
    return all(c in string.hexdigits for c in data)


def get_cipher_list(encryption_mode):
    """
    Get list of ciphers from encryption mode.

    Example:
    get_cipher_list("psk2+tkip+aes") -> ["TKIP", "CCMP"]
    """
    parts = encryption_mode.lower().split('+')
    ciphers = []

    if "tkip" in parts:
        ciphers.append("TKIP")
    if "ccmp" in parts or "aes" in parts:
        ciphers.append("CCMP")

    if len(ciphers) == 0:
        # We need to enable at least one cipher. Most modes default to CCMP
        # except for wpa.
        if parts[0] == "wpa":
            ciphers.append("TKIP")
        else:
            ciphers.append("CCMP")

    return ciphers


class ConfigWifiDevice(ConfigObject):
    typename = "wifi-device"

    options = [
        ConfigOption(name="type", required=True),
        ConfigOption(name="phy", type=str),
        ConfigOption(name="macaddr", type=str),
        ConfigOption(name="ifname", type=str),
        ConfigOption(name="channel", type=int, required=True),
        ConfigOption(name="hwmode"),
        ConfigOption(name="txpower", type=int),
        ConfigOption(name="country"),
        ConfigOption(name="require_mode"),
        ConfigOption(name="htmode"),
        ConfigOption(name="beacon_int", type=int),
        ConfigOption(name="frag", type=int),
        ConfigOption(name="rts", type=int),

        # 802.11n Capabilities
        ConfigOption(name="ldpc", type=bool),
        ConfigOption(name="short_gi_20", type=bool),
        ConfigOption(name="short_gi_40", type=bool),
        ConfigOption(name="tx_stbc", type=int),
        ConfigOption(name="rx_stbc", type=int),
        ConfigOption(name="max_amsdu", type=bool),
        ConfigOption(name="dsss_cck_40", type=bool),

        # 802.11ac Capabilities
        ConfigOption(name="rxldpc", type=bool),
        ConfigOption(name="short_gi_80", type=bool),
        ConfigOption(name="short_gi_160", type=bool),
        ConfigOption(name="tx_stbc_2by1", type=bool),
        ConfigOption(name="rx_antenna_pattern", type=bool),
        ConfigOption(name="tx_antenna_pattern", type=bool),
        ConfigOption(name="vht_max_mpdu", type=int),
        ConfigOption(name="rx_stbc", type=int)
    ]

    def detectPrimaryInterface(self):
        """
        Find the primary network interface associated with this Wi-Fi device.

        By primary we mean the first interface (e.g. wlan0 or wlan1) that
        exists at system startup before any `interface add` commands.  We will
        use the primary interface first, and create additional virtual
        interfaces after that.

        That seems overly complicated, but it is required in cases where the
        Wi-Fi device does not support virtual interfaces.

        Returns interface name or None.
        """
        matches = []

        # Search through all network interfaces.  This includes non-wireless
        # interfaces, so we will get some exceptions trying to read from
        # the phy80211 subdirectory.
        for ifname in os.listdir('/sys/class/net'):
            try:
                path = '/sys/class/net/{}/ifindex'.format(ifname)
                with open(path, 'r') as source:
                    ifindex = int(source.read().rstrip())

                info = {
                    'ifname': ifname,
                    'ifindex': ifindex
                }

                # Check if the current interface matches by phy name.
                path = '/sys/class/net/{}/phy80211/name'.format(ifname)
                with open(path, 'r') as source:
                    phy = source.read().rstrip()
                if phy == self.phy:
                    matches.append(info)
                    continue

                # Check if the current interface matches by macaddress.
                path = '/sys/class/net/{}/phy80211/macaddress'.format(ifname)
                with open(path, 'r') as source:
                    macaddr = source.read().rstrip()
                if macaddr == self.macaddr:
                    matches.append(info)
                    continue
            except:
                pass

        if len(matches) > 0:
            # Sort by ifindex - lower ifindex means it was created earlier.
            matches.sort(key=operator.itemgetter('ifindex'))
            return matches[0]['ifname']
        else:
            return None

    def nextInterfaceName(self):
        """
        Get the next available interface name.
        """
        if self._primary_available:
            ifname = self._ifname
            self._primary_available = False
        else:
            ifname = "v{}.{:04x}".format(self._ifname, self._ifname_counter)
            self._ifname_counter += 1
        return ifname

    def releaseInterfaceName(self, ifname):
        """
        Mark an interface name as no longer used.
        """
        if ifname == self._ifname:
            self._primary_available = True

    def setup(self):
        self._phy = self.phy
        if self._phy is None and self.macaddr is not None:
            self._phy = getPhyFromMAC(self.macaddr)

        # Used to generate unique names for virtual interfaces.
        self._ifname_counter = 0
        self._primary_available = True

        self._ifname = self.ifname
        if self._ifname is None:
            self._ifname = self.detectPrimaryInterface()
        if self._ifname is None:
            #TODO put a random hex string here.
            self._ifname = "unknown"
            self._primary_available = False

    def apply(self, allConfigs):
        commands = []

        # On some systems the primary interface (e.g. wlan0) starts in the UP
        # state and in managed mode.  This would normally be fine, but for some
        # devices, notably the WLE600VX, the driver/firmare throws a "Device or
        # resource busy" error if both a managed mode and an __ap mode
        # interface are UP.
        #
        # To make everyone happy, put the primary interface DOWN while we
        # initialize the device, and if any wifi-iface sections are using it,
        # they will bring it back UP.
        primary = self.detectPrimaryInterface()
        if primary is not None:
            cmd = ["ip", "link", "set", "dev", primary, "down"]
            commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

        if self.txpower is not None:
            # txpower is in dBm, iw takes mBm.
            power = self.txpower * 100
            cmd = ["iw", "phy", self._phy, "set", "txpower",
                    "fixed", str(power)]
        else:
            cmd = ["iw", "phy", self._phy, "set", "txpower", "auto"]

        commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))
        return commands

    def revert(self, allConfigs):
        commands = []

        # We only need to revert txpower if it was actually set in the apply
        # step.
        if self.txpower is not None:
            cmd = ["iw", "phy", self._phy, "set", "txpower", "auto"]
            commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        # Bring primary interface back UP - see comment in apply function.
        primary = self.detectPrimaryInterface()
        if primary is not None:
            cmd = ["ip", "link", "set", "dev", primary, "up"]
            commands.append((-self.PRIO_CREATE_IFACE, Command(cmd, self)))

        return commands


class ConfigWifiIface(ConfigObject):
    typename = "wifi-iface"

    options = [
        ConfigOption(name="device", required=True),
        ConfigOption(name="mode", required=True),
        ConfigOption(name="ssid", default="Paradrop"),
        ConfigOption(name="hidden", type=bool, default=False),
        ConfigOption(name="isolate", type=bool, default=False),
        ConfigOption(name="wmm", type=bool, default=True),
        ConfigOption(name="network", default="lan"),
        ConfigOption(name="encryption", default="none"),
        ConfigOption(name="key"),
        ConfigOption(name="maxassoc", type=int),

        ConfigOption(name="doth", type=bool, default=True),
        ConfigOption(name="short_preamble", type=bool, default=True),

        # WPA Enterprise (RADIUS) options
        ConfigOption(name="auth_server"),
        ConfigOption(name="auth_port", type=int, default=1812),
        ConfigOption(name="auth_secret"),
        ConfigOption(name="acct_server"),
        ConfigOption(name="acct_port", type=int, default=1813),
        ConfigOption(name="acct_secret"),
        ConfigOption(name="nasid"),
        ConfigOption(name="ownip"),
        ConfigOption(name="dynamic_vlan", type=int, default=0),

        # WPA Enterprise (Client)
        ConfigOption(name="identity"),
        ConfigOption(name="password"),

        # 802.11r - Fast BSS Transition
        ConfigOption(name="ieee80211r", type=bool, default=False),
        ConfigOption(name="mobility_domain", default="4f57"),
        ConfigOption(name="r0_key_lifetime", type=int, default=10000),
        ConfigOption(name="r1_key_holder", default="00004f577274"),
        ConfigOption(name="reassociation_deadline", type=int, default=1000),
        ConfigOption(name="r0kh", type=list, default=[]),
        ConfigOption(name="r1kh", type=list, default=[]),
        ConfigOption(name="pmk_r1_push", type=bool, default=False),

        # Nonstandard: interval (seconds) for sending interim updates.
        ConfigOption(name="acct_interval", type=int, default=300),

        # NOTE: ifname is not defined in the UCI specs.  We use it to declare a
        # desired name for the virtual wireless interface that should be
        # created.
        #
        # DEPRECATED: Remove in 1.0 release.
        ConfigOption(name="ifname")
    ]

    def getIfname(self, device, interface):
        """
        Returns the name to be used by this WiFi interface, e.g. as seen by
        ifconfig.

        This comes from the "ifname" option if it is set.  Otherwise, we use
        the interface name of the associated network.
        """
        if self.ifname is not None:
            # We were given a specific ifname to use (deprecated).
            return self.ifname
        elif interface.type == "bridge":
            # It's a bridge (e.g. br-lan), so generate a name based on the
            # device properties.
            return device.nextInterfaceName()
        else:
            # It's not a bridge (e.g. a chute's network), so use the the ifname
            # from the network interface section.
            return interface.config_ifname

    def getName(self):
        """
        Return a unique and consistent identifier for the section.

        If ifname is set, then that is a good choice for the name because
        interface names need to be unique on the system.

        If ifname is not set, then we use the combined string device:network.
        The assumption is that no one will put multiple APs on the same device
        and same network, or if they do, (e.g. multiple APs on the br-lan
        bridge), then they will configure the ifname to be unique.
        """
        if self.ifname is not None:
            return self.ifname
        else:
            return "{}:{}".format(self.device, self.network)

    def apply(self, allConfigs):
        commands = list()

        # Look up the wifi-device section.
        wifiDevice = self.lookup(allConfigs, "wireless", "wifi-device", self.device)
        self._device = wifiDevice

        # Look up the interface section.
        interface = self.lookup(allConfigs, "network", "interface", self.network)

        # Type to pass to iw command, e.g. '__ap'.
        iw_type = IW_IFACE_TYPE[self.mode]

        # Set _ifname while the section is being applied so that it is still
        # available when the section is being reverted.
        self._ifname = self.getIfname(wifiDevice, interface)

        # If this is a primary interface, it should always exist.  We just need
        # to change its mode.  Otherwise, it is a virtual interface, and we need
        # to create it.
        if self._ifname == wifiDevice._ifname:
            # It needs to be "down" before changing type.
            cmd = ["ip", "link", "set", "dev", self._ifname, "down"]
            commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

            cmd = ["iw", "dev", self._ifname, "set", "type", iw_type]
            commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

        else:
            # Command to create the virtual interface.
            cmd = ["iw", "phy", wifiDevice._phy, "interface", "add",
                   self._ifname, "type", iw_type]
            commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

            # Assign a random MAC address to avoid conflict with other
            # interfaces using the same device.
            cmd = ["ip", "link", "set", "dev", self._ifname,
                    "address", self.getRandomMAC()]
            commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

        if self.mode == "ap":
            confFile = self.makeHostapdConf(wifiDevice, interface)

            self.pidFile = "{}/hostapd-{}.pid".format(
                self.manager.writeDir, self.internalName)

            cmd = ["hostapd", "-P", self.pidFile, "-B", confFile]
            commands.append((self.PRIO_START_DAEMON, Command(cmd, self)))

        elif self.mode == "sta":
            confFile = self.makeWpaSupplicantConf(wifiDevice, interface)

            self.pidFile = "{}/wpa_supplicant-{}.pid".format(
                self.manager.writeDir, self.internalName)

            cmd = ["wpa_supplicant", "-B", "-c", confFile, "-i", self._ifname,
                    "-D", "nl80211", "-P", self.pidFile]
            commands.append((self.PRIO_START_DAEMON, Command(cmd, self)))

        return commands

    def makeHostapdConf(self, wifiDevice, interface):
        outputPath = "{}/hostapd-{}.conf".format(
            self.manager.writeDir, self.internalName)

        conf = HostapdConfGenerator(self, wifiDevice, interface)
        conf.generate(outputPath)

        return outputPath

    def makeWpaSupplicantConf(self, wifiDevice, interface):
        outputPath = "{}/wpa_supplicant-{}.conf".format(
            self.manager.writeDir, self.internalName)

        conf = WpaSupplicantConfGenerator(self, wifiDevice, interface)
        conf.generate(outputPath)

        return outputPath

    def revert(self, allConfigs):
        commands = list()

        if self.mode == "ap":
            commands.append((-self.PRIO_START_DAEMON,
                KillCommand(self.pidFile, self)))

        # If this is a primary interface we leave it alone but mark it as
        # available again.  Otherwise, delete the interface.
        if self._ifname != self._device._ifname:
            # Delete our virtual interface.
            cmd = ["iw", "dev", self._ifname, "del"]
            commands.append((-self.PRIO_CREATE_IFACE, Command(cmd, self)))
        self._device.releaseInterfaceName(self._ifname)

        return commands

    def updateApply(self, new, allConfigs):
        if new.mode != self.mode or \
                new.device != self.device or \
                new.network != self.network:
            # Major change requires unloading the old section and applying the
            # new.
            return self.apply(allConfigs)

        commands = list()

        if new.mode == "ap":
            # Look up the wifi-device section.
            wifiDevice = new.lookup(allConfigs, "wireless", "wifi-device", new.device)

            # Look up the interface section.
            interface = new.lookup(allConfigs, "network", "interface", new.network)

            # Copy fields that would normally be set in apply method.
            new._device = wifiDevice
            new._ifname = self._ifname

            confFile = new.makeHostapdConf(wifiDevice, interface)

            new.pidFile = "{}/hostapd-{}.pid".format(
                new.manager.writeDir, new.internalName)

            cmd = ["hostapd", "-P", new.pidFile, "-B", confFile]
            commands.append((self.PRIO_START_DAEMON, Command(cmd, new)))

        return commands

    def updateRevert(self, new, allConfigs):
        if new.mode != self.mode or \
                new.device != self.device or \
                new.network != self.network:
            # Major change requires unloading the old section and applying the
            # new.
            return self.revert(allConfigs)

        commands = list()

        if self.mode == "ap":
            # Bring down hostapd
            commands.append((-self.PRIO_START_DAEMON,
                KillCommand(self.pidFile, self)))

        return commands

    def getRandomMAC(self):
        """
        Generate a random MAC address.

        Returns a string "02:xx:xx:xx:xx:xx".  The first byte is 02, which
        indicates a locally administered address.
        """
        parts = ["02"]
        for i in range(5):
            parts.append("{:02x}".format(random.randrange(0, 255)))
        return ":".join(parts)


class ConfGenerator(object):
    def writeHeader(self, output):
        output.write("#" * 79 + "\n")
        output.write("# Configuration file generated by pdconf\n")
        output.write("#" * 79 + "\n")

    def writeOptions(self, options, output, title=None):
        output.write("\n")
        if title is not None:
            head = "##### {} ".format(title)
            tail = "#" * (79 - len(head))
            output.write(head + tail + "\n")
        for name, value in options:
            if isinstance(value, list):
                output.write("{}={{\n".format(name))
                for subname, subvalue, quote in value:
                    if quote:
                        output.write("\t{}=\"{}\"\n".format(subname, subvalue))
                    else:
                        output.write("\t{}={}\n".format(subname, subvalue))
                output.write("}\n")
            else:
                output.write("{}={}\n".format(name, value))


class HostapdConfGenerator(ConfGenerator):
    def __init__(self, wifiIface, wifiDevice, interface):
        self.wifiIface = wifiIface
        self.wifiDevice = wifiDevice
        self.interface = interface

        self.readMode(wifiDevice)

    def readMode(self, device):
        """
        Determine HT/VHT mode if applicable.
        """
        self.enable11n = False
        self.enable11ac = False

        if device.htmode is None:
            return

        if device.htmode.startswith("HT"):
            self.enable11n = True
        elif device.htmode.startswith("VHT"):
            self.enable11n = True
            self.enable11ac = True

    def generate(self, path):
        with open(path, "w") as output:
            self.writeHeader(output)

            options = self.getMainOptions()
            self.writeOptions(options, output)

            if self.enable11n:
                options = self.get11nOptions()
                self.writeOptions(options, output, title="802.11n")

            if self.enable11ac:
                options = self.get11acOptions()
                self.writeOptions(options, output, title="802.11ac")

            options = self.getSecurityOptions()
            self.writeOptions(options, output, title="Security")

            options = self.getRadiusOptions()
            if len(options) > 0:
                self.writeOptions(options, output, title="RADIUS Client")

            if self.wifiIface.ieee80211r:
                options = self.get11rOptions()
                self.writeOptions(options, output, title="802.11r")

    def getMainOptions(self):
        options = list()

        options.append(("interface", self.wifiIface._ifname))

        if self.interface.type == "bridge":
            options.append(("bridge", self.interface.config_ifname))

        control_dir = os.path.join(self.wifiIface.manager.writeDir, "hostapd")
        options.append(("ctrl_interface", control_dir))

        # Do not set ctrl_interface_group because it triggers behavior
        # that breaks under snap confinement.
        #options.append(("ctrl_interface_group", 0))

        options.append(("ssid", self.wifiIface.ssid))

        if self.wifiDevice.country is not None:
            options.append(("country_code", self.wifiDevice.country))
            options.append(("ieee80211d", 1))
            if self.wifiIface.doth:
                options.append(("ieee80211h", 1))

        hwmode = self.wifiDevice.hwmode
        if hwmode is not None:
            if hwmode in HOSTAPD_HWMODE:
                # Convert UCI hwmode to hostapd hwmode format.
                options.append(("hw_mode", HOSTAPD_HWMODE[hwmode]))
            else:
                raise Exception("Unrecognized hardware mode: {}".format(hwmode))

        options.append(("channel", self.wifiDevice.channel))

        if self.wifiDevice.beacon_int is not None:
            options.append(("beacon_int", self.wifiDevice.beacon_int))

        if self.wifiIface.maxassoc is not None:
            options.append(("max_num_sta", self.wifiIface.maxassoc))
            options.append(("no_probe_resp_if_max_sta", 1))

        # 1 = send empty (length=0) SSID in beacon and ignore probe request for
        #     broadcast SSID
        # 2 = clear SSID (ASCII 0), but keep the original length (this may be required
        #     with some clients that do not support empty SSID) and ignore probe
        #     requests for broadcast SSID
        if self.wifiIface.hidden:
            options.append(("ignore_broadcast_ssid", 2))

        if self.wifiIface.isolate:
            options.append(("ap_isolate", 1))

        if self.wifiDevice.rts is not None:
            options.append(("rts_threshold", self.wifiDevice.rts))
        if self.wifiDevice.frag is not None:
            options.append(("fragm_threshold", self.wifiDevice.rts))

        if self.wifiIface.short_preamble:
            options.append(("preamble", 1))

        options.append(("wmm_enabled", 1 * self.wifiIface.wmm))

        return options

    def get11nOptions(self):
        options = list()

        options.append(("ieee80211n", 1 * self.enable11n))

        ht_capab = ""
        if self.wifiDevice.htmode.startswith("HT40"):
            ht_capab += "[{}]".format(self.wifiDevice.htmode)
        elif self.wifiDevice.htmode in ["VHT40", "VHT80", "VHT160"]:
            if self.wifiDevice.channel in HT40_LOWER_CHANNELS:
                ht_capab += "[HT40+]"
            elif self.wifiDevice.channel in HT40_UPPER_CHANNELS:
                ht_capab += "[HT40-]"

        if self.wifiDevice.ldpc:
            ht_capab += "[LDPC]"
        if self.wifiDevice.short_gi_20:
            ht_capab += "[SHORT-GI-20]"
        if self.wifiDevice.short_gi_40:
            ht_capab += "[SHORT-GI-40]"
        if self.wifiDevice.tx_stbc:
            ht_capab += "[TX-STBC]"
        if self.wifiDevice.rx_stbc == 1:
            ht_capab += "[RX-STBC1]"
        elif self.wifiDevice.rx_stbc == 2:
            ht_capab += "[RX-STBC12]"
        elif self.wifiDevice.rx_stbc >= 3:
            ht_capab += "[RX-STBC123]"
        if self.wifiDevice.max_amsdu:
            ht_capab += "[MAX-AMSDU-7935]"
        if self.wifiDevice.dsss_cck_40:
            ht_capab += "[DSSS_CCK-40]"

        if len(ht_capab) > 0:
            options.append(("ht_capab", ht_capab))

        if self.wifiDevice.require_mode == "n":
            options.append(("require_ht", 1))

        return options

    def get11acOptions(self):
        options = list()

        options.append(("ieee80211ac", 1 * self.enable11ac))

        if self.wifiDevice.require_mode == "ac":
            options.append(("require_vht", 1))

        # Default chwidth=0 means 20 or 40 Mhz.
        chwidth = 0
        seg0_idx = self.wifiDevice.channel
        seg1_idx = None
        if self.wifiDevice.htmode == "VHT40":
            seg0_idx = VHT40_CENTER_INDEX[self.wifiDevice.channel]
        elif self.wifiDevice.htmode == "VHT80":
            chwidth = 1
            seg0_idx = VHT80_CENTER_INDEX[self.wifiDevice.channel]
        elif self.wifiDevice.htmode == "VHT160":
            chwidth = 2
            seg0_idx = VHT160_CENTER_INDEX[self.wifiDevice.channel]
        # TODO: How does the admin request 80+80 mode (chwidth=3)?
        # We need a second channel number to specify seg1_idx in that case.

        vht_capab = ""
        if self.wifiDevice.rxldpc:
            vht_capab += "[RXLDPC]"
        if self.wifiDevice.short_gi_80:
            vht_capab += "[SHORT-GI-80]"
        if self.wifiDevice.short_gi_160:
            vht_capab += "[SHORT-GI-160]"
        if self.wifiDevice.tx_stbc_2by1:
            vht_capab += "[TX-STBC-2BY1]"
        if self.wifiDevice.rx_antenna_pattern:
            vht_capab += "[RX-ANTENNA-PATTERN]"
        if self.wifiDevice.tx_antenna_pattern:
            vht_capab += "[TX-ANTENNA-PATTERN]"
        if self.wifiDevice.vht_max_mpdu == 7991:
            vht_capab += "[MAX-MPDU-7991]"
        elif self.wifiDevice.vht_max_mpdu == 11454:
            vht_capab += "[MAX-MPDU-11454]"
        if self.wifiDevice.rx_stbc == 1:
            vht_capab += "[RX-STBC-1]"
        elif self.wifiDevice.rx_stbc == 2:
            vht_capab += "[RX-STBC-12]"
        elif self.wifiDevice.rx_stbc == 3:
            vht_capab += "[RX-STBC-123]"
        elif self.wifiDevice.rx_stbc >= 4:
            vht_capab += "[RX-STBC-1234]"
        if len(vht_capab) > 0:
            options.append(("vht_capab", vht_capab))

        options.append(("vht_oper_chwidth", chwidth))
        options.append(("vht_oper_centr_freq_seg0_idx", seg0_idx))
        if seg1_idx is not None:
            options.append(("vht_oper_centr_freq_seg1_idx", seg1_idx))

        return options

    def getSecurityOptions(self):
        options = list()

        if self.wifiIface.encryption is None or \
                self.wifiIface.encryption == "none":
            options.append(("wpa", 0))
            return options

        modes = self.wifiIface.encryption.split("+")

        # Check for WPA, WPA2, or mixed mode.
        if modes[0] in ["psk2", "wpa2"]:
            options.append(("wpa", 2))
        elif modes[0] in ["psk", "wpa"]:
            options.append(("wpa", 1))
        elif modes[0] in ["psk-mixed", "wpa-mixed"]:
            options.append(("wpa", 3))
        else:
            raise Exception("Encryption mode ({}) not supported".format(modes[0]))

        # Check for PSK vs Enterprise mode.
        if modes[0] in ["psk2", "psk", "psk-mixed"]:
            # If key is a 64 character hex string, then treat it as the PSK
            # directly, else treat it as a passphrase.
            if len(self.wifiIface.key) == 64 and isHexString(self.wifiIface.key):
                options.append(("wpa_psk", self.wifiIface.key))
            else:
                options.append(("wpa_passphrase", self.wifiIface.key))

            options.append(("wpa_key_mgmt", "WPA-PSK"))
        elif modes[0] in ["wpa2", "wpa", "wpa-mixed"]:
            if self.wifiIface.auth_server is None:
                raise Exception("Must configure auth_server with wpa2")

            options.append(("wpa_key_mgmt", "WPA-EAP"))

        ciphers = get_cipher_list(self.wifiIface.encryption)

        # Ciphers to use for WPA:
        options.append(("wpa_pairwise", " ".join(ciphers)))

        # Ciphers to use for WPA2:
        # Technically, these can be different, e.g. wpa_pairwise=TKIP and
        # rsn_pairwise=CCMP. I am not sure we need to expose those options.
        # Defaulting to CCMP only is good because TKIP is no longer secure.
        options.append(("rsn_pairwise", " ".join(ciphers)))

        return options

    def getRadiusOptions(self):
        options = list()

        if self.wifiIface.ownip:
            options.append(("own_ip_addr", self.wifiIface.ownip))
        if self.wifiIface.nasid:
            options.append(("nas_identifier", self.wifiIface.nasid))

        if self.wifiIface.auth_server:
            options.append(("auth_server_addr", self.wifiIface.auth_server))
            options.append(("auth_server_port", self.wifiIface.auth_port))
            if self.wifiIface.auth_secret:
                options.append(("auth_server_shared_secret",
                    self.wifiIface.auth_secret))
            options.append(("dynamic_vlan", self.wifiIface.dynamic_vlan))

        if self.wifiIface.acct_server:
            options.append(("acct_server_addr", self.wifiIface.acct_server))
            options.append(("acct_server_port", self.wifiIface.acct_port))
            if self.wifiIface.acct_secret:
                options.append(("acct_server_shared_secret",
                    self.wifiIface.acct_secret))
            options.append(("radius_acct_interim_interval", self.wifiIface.acct_interval))

        return options

    def get11rOptions(self):
        """
        Get options related to 802.11r (fast BSS transition).
        """
        options = list()

        if self.wifiIface.nasid is None:
            raise Exception("nasid is not defined but is required for 802.11r")

        options.append(("mobility_domain", self.wifiIface.mobility_domain))
        options.append(("r0_key_lifetime", self.wifiIface.r0_key_lifetime))
        options.append(("r1_key_holder", self.wifiIface.r1_key_holder))
        options.append(("reassociation_deadline", self.wifiIface.reassociation_deadline))

        for item in self.wifiIface.r0kh:
            # May be specified as either comma- or space-separated, but hostapd
            # expects spaces.
            parts = re.split("[, ]+", item)
            options.append(("r0kh", " ".join(parts)))

        for item in self.wifiIface.r1kh:
            # May be specified as either comma- or space-separated, but hostapd
            # expects spaces.
            parts = re.split("[, ]+", item)
            options.append(("r1kh", " ".join(parts)))

        return options

    def writeHeader(self, output):
        output.write("#" * 79 + "\n")
        output.write("# hostapd configuration file generated by pdconf\n")
        output.write("# Source: {}\n".format(self.wifiIface.source))
        output.write("# Section: {}\n".format(str(self.wifiIface)))
        output.write("# Device: {}\n".format(str(self.wifiDevice)))
        output.write("# Interface: {}\n".format(str(self.interface)))
        output.write("#" * 79 + "\n")


class WpaSupplicantConfGenerator(ConfGenerator):
    def __init__(self, wifiIface, wifiDevice, interface):
        self.wifiIface = wifiIface
        self.wifiDevice = wifiDevice
        self.interface = interface

    def generate(self, path):
        with open(path, "w") as output:
            self.writeHeader(output)

            options = self.getMainOptions()
            self.writeOptions(options, output)

    def getMainOptions(self):
        options = list()

        options.append(("ap_scan", 1))
        if self.wifiDevice.country is not None:
            options.append(("country", self.wifiDevice.country))

        # (name, value, quote)
        network = list()

        network.append(('ssid', self.wifiIface.ssid, True))
        network.append(('scan_ssid', 1, False))

        modes = self.wifiIface.encryption.split("+")
        if modes[0] == "psk2":
            network.append(('key_mgmt', 'WPA-PSK', False))
            network.append(('psk', self.wifiIface.key, True))

        elif modes[0] == "wpa2":
            network.append(('key_mgmt', 'WPA-EAP', False))
            network.append(('identity', self.wifiIface.identity))
            network.append(('password', self.wifiIface.password))

        else:
            network.append(('key_mgmt', 'NONE', False))

        options.append(("network", network))

        return options

    def writeHeader(self, output):
        output.write("#" * 79 + "\n")
        output.write("# wpa_supplicant configuration file generated by pdconf\n")
        output.write("# Source: {}\n".format(self.wifiIface.source))
        output.write("# Section: {}\n".format(str(self.wifiIface)))
        output.write("# Device: {}\n".format(str(self.wifiDevice)))
        output.write("# Interface: {}\n".format(str(self.interface)))
        output.write("#" * 79 + "\n")



import binascii
import jwt
import os
import time

from paradrop.base import nexus


def generate_jwt_secret():
    # 32 bytes = 256 bits, recommended key length for HS256.
    material = os.urandom(32)
    return binascii.b2a_hex(material)


class TokenManager(object):
    # Inherit this exception class so that callers can use it without depending
    # on the jwt library.
    InvalidTokenError = jwt.exceptions.InvalidTokenError

    def __init__(self):
        self.secret = nexus.core.getKey('jwt-secret')
        if self.secret is None:
            self.secret = generate_jwt_secret()
            nexus.core.saveKey(self.secret, 'jwt-secret')

    def decode(self, token):
        """
        Decode a JWT that was issued by us.

        Throws an InvalidTokenError on decoding failure or token expiration.
        """
        return jwt.decode(token, self.secret, algorithms=['HS256'])

    def issue(self, subject, expires=2592000, **claims):
        """
        Issue a signed token.

        The subject is mandatory and sets the JWT "sub" claim.  The expiration
        time can be None (no expiration) or a time in seconds from the issue
        time after which the token will expire.  The default is 30 days
        (2592000 seconds).
        """
        data = claims
        data['iat'] = int(time.time())
        data['iss'] = nexus.core.info.pdid
        data['sub'] = subject
        if expires is not None:
            data['exp'] = int(time.time()) + expires
        return jwt.encode(data, self.secret, algorithm='HS256')



"""
This module is here purely to help with understanding the rather complex
execution plan in Paradrop.  Simply run it (python -m paradrop.plan_demo),
and it will walk through all of the functions that make up a chute creation
operation.
"""
from __future__ import print_function

from paradrop.core.update.update_object import UpdateChute
from paradrop.core.plan import executionplan, plangraph


def loadPriorityMap():
    """
    Make a map of priority values back to their names for reference.

    These are defined as constant integer values in
    paradrop.backend.exc.plangraph.  For example, for priority 9
    (STRUCT_GET_SYSTEM_DEVICES), the dictionary produced by this function
    would contain the entry 9: "STRUCT_GET_SYSTEM_DEVICES".
    """
    priomap = dict()
    for item in dir(plangraph):
        value = getattr(plangraph, item)
        if isinstance(value, int):
            priomap[value] = item
    return priomap


if __name__ == "__main__":
    priomap = loadPriorityMap()

    update = UpdateChute({
        "updateClass": "CHUTE",
        "updateType": "create",
        "name": "test",
        "tok": 0
    })
    executionplan.generatePlans(update)
    executionplan.aggregatePlans(update)

    for prio, func, abort in update.plans.plans:
        print("Module: {}".format(func.func.__module__))
        print("Function: {}".format(func.func.__name__))
        print("Priority: {} ({})".format(priomap[prio], prio))
        if func.func.__doc__ is None:
            print("\n    Function has no docstring.\n")
        else:
            print(func.func.__doc__)




"""
The host configuration controls system settings of the host OS.

This module operates as follows:

1. The first time, we try to detect all devices and auto-generate
a reasonable configuration, which we store to a persistent file.

2. (TODO) We present the configuration to the owner sometime around
provisioning or first chute creation and allow him to change settings.

3. (TODO) We have some kind of update operation that can manipulate
settings.
"""

import ipaddress
import jsonpatch
import yaml

from paradrop.base import settings
from paradrop.lib.utils import pdosq

from . import devices as config_devices

# This patch ensures the "unicode" function exists and behaves consistently
# between Python 2 and 3. Normally, it is recommended to use "builtins.str" for
# compatibility, but that uses a different class type not recognized by PyYAML.
try:
    unicode('')
except NameError:
    unicode = str


CHANNELS_24G = [1, 6, 11]
CHANNELS_5G  = [36, 40, 44, 48, 149, 153, 157, 161, 165]


WIFI_DEVICE_CAPS = {
    # WLE200NX
    ("0x168c", "0x002a"): {
        "hwmode": ["11a", "11g"]
    },

    # WLE600VX
    ("0x168c", "0x003c"): {
        "hwmode": ["11a", "11g"]
    },

    "default": {
        "hwmode": ["11g"]
    }
}

WIFI_DEVICE_PROFILE = {
    # WLE200NX
    ("0x168c", "0x002a"): {
        "htmode": "HT20",
        "tx_stbc": 1,
        "rx_stbc": 1,
        "short_gi_40": True,
    },

    # WLE600VX
    ("0x168c", "0x003c"): {
        "htmode": "VHT20",
        "tx_stbc": 1,
        "rx_stbc": 1,
        "short_gi_20": True,
        "short_gi_40": True,
        "short_gi_80": True
    },

    "default": {}
}


def save(config, path=None):
    """
    Save host configuration.

    May raise exception if unable to write the configuration file.
    """
    if path is None:
        path = settings.HOST_CONFIG_FILE

    with open(path, 'w') as output:
        output.write(yaml.safe_dump(config, default_flow_style=False))


def load(path=None):
    """
    Load host configuration.

    Tries to load host configuration from persistent file.  If that does not
    work, it will try to automatically generate a working configuration.

    Returns a host config object on success or None on failure.
    """
    if path is None:
        path = settings.HOST_CONFIG_FILE

    return pdosq.read_yaml_file(path, default=None)


def generateHostConfig(devices):
    """
    Scan for devices on the machine and generate a working configuration.
    """
    config = dict()

    default_lan_network = ipaddress.ip_network(unicode(settings.DEFAULT_LAN_NETWORK))

    config['firewall'] = {
        'defaults': {
            'input': 'ACCEPT',
            'output': 'ACCEPT',
            'forward': 'ACCEPT'
        }
    }
    config['lan'] = {
        'interfaces': list(),
        'proto': 'static',
        'ipaddr': settings.DEFAULT_LAN_ADDRESS,
        'netmask': unicode(default_lan_network.netmask),
        'dhcp': {
            'start': 100,
            'limit': 100,
            'leasetime': '12h'
        },
        'firewall': {
            'defaults': {
                'conntrack': '1',
                'input': 'ACCEPT',
                'output': 'ACCEPT',
                'forward': 'ACCEPT'
            },
            'forwarding': []
        }
    }
    config['wifi'] = list()
    config['wifi-interfaces'] = list()
    config['system'] = {
        'autoUpdate': True,
        'chuteSubnetPool': settings.DYNAMIC_NETWORK_POOL,
        'chutePrefixSize': 24,
        'onMissingWiFi': None
    }
    config['telemetry'] = {
        'enabled': True,
        'interval': 60
    }
    config['zerotier'] = {
        'enabled': True,
        'networks': []
    }

    if len(devices['wan']) > 0:
        wanDev = devices['wan'][0]
        config['wan'] = {
            'interface': wanDev['name'],
            'proto': 'dhcp',
            'firewall': {
                'defaults': {
                    'conntrack': '1',
                    'masq': '1',
                    'input': 'ACCEPT',
                    'output': 'ACCEPT',
                    'forward': 'ACCEPT',
                    'masq_src': [
                        settings.DEFAULT_LAN_NETWORK,
                        settings.DYNAMIC_NETWORK_POOL
                    ]
                }
            }
        }

        # Add a rule that forwards LAN traffic to WAN.
        config['lan']['firewall']['forwarding'].append({
            'src': 'lan',
            'dest': 'wan'
        })

    for lanDev in devices['lan']:
        config['lan']['interfaces'].append(lanDev['name'])

    # Counters to help with channel assignment.
    serial = config_devices.get_hardware_serial()
    wifi_24_assigned = 0
    wifi_5_assigned = 0

    for wifiDev in devices['wifi']:
        pair = (wifiDev['vendor'], wifiDev['device'])
        if pair in WIFI_DEVICE_PROFILE:
            new_config = WIFI_DEVICE_PROFILE[pair].copy()
            wifi_caps = WIFI_DEVICE_CAPS[pair]
        else:
            new_config = WIFI_DEVICE_PROFILE['default'].copy()
            wifi_caps = WIFI_DEVICE_CAPS["default"]

        # Skip odd devices that are missing this field.
        if 'id' not in wifiDev:
            continue

        new_config['id'] = wifiDev['id']

        # This logic will assign a 5 Ghz channel to the first 5 Ghz-capable
        # device.  It will then alternate between 2.4 Ghz and 5 Ghz.
        choose_5g = "11a" in wifi_caps.get('hwmode', []) and \
                wifi_5_assigned <= wifi_24_assigned

        if choose_5g:
            chan_index = (serial + wifi_5_assigned) % len(CHANNELS_5G)
            new_config['channel'] = CHANNELS_5G[chan_index]
            new_config['hwmode'] = "11a"
            wifi_5_assigned += 1
        else:
            chan_index = (serial + wifi_24_assigned) % len(CHANNELS_24G)
            new_config['channel'] = CHANNELS_24G[chan_index]
            new_config['hwmode'] = "11g"
            wifi_24_assigned += 1

        config['wifi'].append(new_config)

    if len(config['wifi']) > 0 and settings.DEFAULT_WIRELESS_ENABLED:
        # If we detect WiFi devices now, configure the system to warn if they
        # are missing later.  Production systems should be configured with
        # "reboot".
        config['system']['onMissingWiFi'] = "warn"

        # Add a default WiFi AP for usability.
        new_iface = {
            'device': devices['wifi'][0]['id'],
            'ssid': settings.DEFAULT_WIRELESS_ESSID,
            'mode': 'ap',
            'network': 'lan'
        }

        if settings.DEFAULT_WIRELESS_KEY:
            new_iface['encryption'] = "psk2"
            new_iface['key'] = settings.DEFAULT_WIRELESS_KEY

        config['wifi-interfaces'].append(new_iface)

    return config


def prepareHostConfig(devices=None, hostConfigPath=None, write=True):
    """
    Load an existing host configuration or generate one.

    Tries to load host configuration from persistent file.  If that does not
    work, it will try to automatically generate a working configuration.

    write: if True and host config was automatically generated, then write
    the new host config to a file.
    """
    config = load(hostConfigPath)
    if config is not None:
        return config

    if devices is None:
        devices = config_devices.detectSystemDevices()
    config = generateHostConfig(devices)

    if write:
        save(config)

    return config


#
# Chute update operations
#


def getHostConfig(update):
    """
    Load host configuration.

    Read device information from networkDevices.
    Store host configuration in hostConfig.
    """
    # TODO We need to check for changes in hardware.  If a new device was
    # added, we should try to automatically configure it.  If a device was
    # removed, we should be aware of what is no longer valid.
    devices = update.cache_get('networkDevices')
    config = prepareHostConfig(devices)

    # update.old is not guaranteed to contain the old host configuration, so
    # save a backup copy in update.new.  This will be used by revertHostConfig
    # if we need to back out.
    update.cache_set('oldHostConfig', config)

    # If this is a sethostconfig operation, then read the host config from the
    # update object.  Ordinary chute operations should not alter the host
    # configuration.
    if update.updateType == 'sethostconfig':
        config = update.hostconfig

    elif update.updateType == 'patchhostconfig':
        config = jsonpatch.apply_patch(config, update.patch)

    # For factoryreset, try to load the default configuration or automatically
    # generate a new one if the file is not found.
    elif update.updateType == 'factoryreset':
        config = prepareHostConfig(devices,
                hostConfigPath=settings.DEFAULT_HOST_CONFIG_FILE)

    update.cache_set('hostConfig', config)


def setHostConfig(update):
    """
    Write host configuration to persistent storage.

    Read host configuration from hostConfig.
    """
    config = update.cache_get('hostConfig')
    save(config)


def revertHostConfig(update):
    """
    Restore host configuration from before update.

    Uses oldHostConfig cache entry.
    """
    config = update.cache_get('oldHostConfig')
    save(config)



import os
import crypt
import random

from paradrop.base import settings
from paradrop.lib.utils import pdos

class PasswordManager(object):
    def __init__(self):
        self.password_file = os.path.join(settings.CONFIG_HOME_DIR, 'password')

        # Try to parse the password file
        # self.records will have the pairs of user name and password hash
        parsed = False
        if pdos.exists(self.password_file):
            lines = pdos.readFile(self.password_file)
            if lines:
                for line in lines:
                    elements = line.split(':')
                    if len(elements) == 2:
                        self.records = []
                        self.records.append({
                            'user_name': elements[0],
                            'password_hash': elements[1]
                        })
                if len(lines) == len(self.records):
                    parsed = True
                else:
                    self.records = []

        if not parsed:
            self.reset()

    def _sync_password_file(self):
        file_content = ''
        for record in self.records:
            file_content = file_content + record['user_name'] + ':' + record['password_hash'] + '\n'

        pdos.write(self.password_file, file_content)

    def _generate_salt(self):
        # The salt can be generated with crypt.mksalt() on Python 3
        CHARACTERS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        salt = ''.join(random.choice(CHARACTERS) for i in range(16))
        # Use SHA512
        return '$6$' + salt

    def _retrieve_salt(self, password_hash):
        # Example hashed password
        # python -c 'import crypt; print crypt.crypt("", "$6$cMLi2vqIIJtq1Shm")'
        # $6$cMLi2vqIIJtq1Shm$41Ko1W2aTapn.p12G2dWRrRdStI2CrkC1JsftC/bPmwVgiy0vAQIXuuZEbVenyqsM6vZOsuzFrKrQC9NGd6/p.
        elements = password_hash.split('$')

        if len(elements) != 4:
            return None
        else:
            # Use SHA512
            return '$6$' + elements[2]

    def _hash_password(self, password):
        salt = self._generate_salt()
        return crypt.crypt(password, salt)

    def reset(self):
        self.records = []
        self.records.append({
            'user_name': settings.DEFAULT_PANEL_USERNAME,
            'password_hash': self._hash_password(settings.DEFAULT_PANEL_PASSWORD)
        })

        self._sync_password_file()

    def add_user(self, user_name, password):
        found_records = filter(lambda x: x['user_name'] == user_name, self.records)
        count = len(found_records)

        if count == 0:
            self.records.append({
                'user_name': user_name,
                'password_hash': self._hash_password(password)
            })
            self._sync_password_file()
            return True
        else:
            return False

    def remove_user(self, user_name):
        origin_len = len(self.records)
        self.records = [x for x in self.records if x['user_name'] != user_name]

        if len(self.records) != origin_len:
            self._sync_password_file()

    def verify_password(self, user_name, password):
        found_records = filter(lambda x: x['user_name'] == user_name, self.records)
        count = len(found_records)
        if count == 0:
            return False
        elif count == 1:
            password_hash = found_records[0]['password_hash']
            return crypt.crypt(password, self._retrieve_salt(password_hash)) == password_hash
        elif count > 1:
            # Should not be here
            raise Exception('Faint! Is there something wrong?')

    # We need to verify the user with the old password first
    def change_password(self, user_name, newPassword):
        if not user_name:
            user_name = settings.DEFAULT_PANEL_USERNAME

        for i in self.records:
            if i['user_name'] == user_name:
                i['password_hash'] = self._hash_password(newPassword)
                self._sync_password_file()
                return True

        # Could not found the user_name
        return False



import attr
import copy

import six


def interpretBoolean(s):
    """
    Interpret string as a boolean value.

    "0" and "False" are interpreted as False.  All other strings result in
    True.  Technically, only "0" and "1" values should be seen in UCI files.
    However, because of some string conversions in Python, we may encounter
    "False".
    """
    if s in ["0", "False"]:
        return False
    else:
        return True


class ConfigObject(object):
    nextId = 0
    typename = None
    options = []

    # Use to resolve conflicts in with old-style section naming.  maskable=True
    # means a section can be replaced when we are gathering the list of
    # ConfigObject subclasses.  We set maskable=False on network:interface so
    # that it receives precedence over qos:interface and old code continues to
    # work.
    maskable = True

    # Priorities to ensure proper interleaving of commands.
    # Revert commands should use negated value.
    PRIO_CREATE_IFACE = 20
    PRIO_CREATE_VLAN = 25
    PRIO_CONFIG_IFACE = 30
    PRIO_IPTABLES_TOP = 35
    PRIO_IPTABLES_ZONE = 36
    PRIO_IPTABLES_RULE = 37
    PRIO_CREATE_QDISC = 40
    PRIO_CONFIG_QDISC = 45
    PRIO_START_DAEMON = 60

    def __init__(self, name=None):
        self.id = ConfigObject.nextId
        ConfigObject.nextId += 1

        self.epoch = 0
        self.source = None
        self.name = name
        self.comment = None

        self.parents = set()
        self.dependents = set()

        # name is the externally visible name, e.g. "lan" in 
        # "config interface lan".  name is None for anonymous sections,
        # e.g. "config wifi-iface".
        #
        # internalName is used for identifying config objects and must
        # always be defined.  For anonymous sections, we generate a unique
        # name from id.
        if name is None:
            self.internalName = "s{:08x}".format(self.id)
        else:
            self.internalName = name

        # Will be a running list of commands executed by request of this config
        # object.
        self.executed = list()

        for option in self.options:
            setattr(self, option.name, option.default)

    def __hash__(self):
        return hash(self.getTypeAndName())

    def __lt__(self, other):
        """
        Compare ConfigObject instances.

        Currently, it arbitrarily sorts based on the string form of the config
        section type and name (e.g. "config interface wlan0").
        """
        return str(self) < str(other)

    def __str__(self):
        if self.name is None:
            return "config {}".format(self.typename)
        else:
            return "config {} {}".format(self.typename, self.name)

    def setup(self):
        """
        Finish object initialization.

        This is called after the config object is initialized will all of its
        options values filled in.  Override to do some preparation work before
        we start generating commands.
        """
        pass

    def copy(self):
        """
        Make a copy of the config object.

        The copy will receive the same name and option values.
        """
        other = self.__class__()

        other.source = self.source
        other.name = self.name
        other.comment = self.comment

        other.parents = self.parents.copy()
        other.dependents = self.dependents.copy()

        for option in self.options:
            # We use copy here because it works with both the str- and
            # list-typed values.  Any lists are lists of strings, so
            # shallow-copy here is fine.
            copied = copy.copy(getattr(self, option.name))
            setattr(other, option.name, copied)

        # We should call setup on a new config object after all of the option
        # values are filled in.
        other.setup()

        return other

    def dump(self):
        """
        Return full configuration section as a string.
        """
        result = "# internal id: s{:08x}\n".format(self.id)

        if self.name is None:
            result += "config {}".format(self.typename)
        else:
            result += "config {} {}".format(self.typename, self.name)
        if self.comment is not None:
            result += " #" + self.comment
        result += "\n"

        for opdef in self.options:
            value = getattr(self, opdef.name)
            if value is None:
                continue
            elif opdef.type == bool:
                result += "\toption {} '{}'\n".format(opdef.name, 1 * value)
            elif opdef.type == list:
                for v in value:
                    result += "\tlist {} '{}'\n".format(opdef.name, v)
            else:
                result += "\toption {} '{}'\n".format(opdef.name, value)

        return result

    def getName(self):
        """
        Return section name.

        Subclasses that do not have names (anonymous sections) should override
        this to return some other unique identifier such as an interface name.
        """
        return self.internalName

    def getTypeAndName(self):
        """
        Return tuple (section module, section type, section name).
        """
        return (self.getModule(), self.typename, self.getName())

    def lookup(self, allConfigs, sectionModule, sectionType, sectionName, addDependent=True):
        """
        Look up a section by type and name.

        If addDependent is True (default), the current object will be added as
        a dependent of the found section.

        Will raise an exception if the section is not found.
        """
        config = allConfigs[(sectionModule, sectionType, sectionName)]
        if addDependent:
            self.parents.add(config)
            config.dependents.add(self)
        return config

    def removeFromParents(self):
        """
        Remove this section from being tracked by its parents.

        Call this before discarding a configuration section so that later on,
        if the parent is updated, it doesn't try to update non-existent
        children.
        """
        for parent in self.parents:
            if self in parent.dependents:
                parent.dependents.remove(self)
        self.parents.clear()

    def findByType(self, allConfigs, module, typename, where={}):
        """
        Look up sections by type (generator).

        where: filter the returned results by checking option values.
        """
        for key in allConfigs.keys():
            if key[0] == module and key[1] == typename:
                # Skip this section if any of the option values do not match.
                if any(getattr(allConfigs[key], op, None) != where[op] for op in where):
                    continue
                yield allConfigs[key]

    def optionsMatch(self, other):
        """
        Test equality of config sections by comparing option values.
        """
        if not isinstance(other, self.__class__):
            return False
        for opdef in self.options:
            if getattr(self, opdef.name) != getattr(other, opdef.name):
                return False
        return True

    #
    # The following methods (apply, revert, updateApply, and updateRevert)
    # are the most important for subclasses to override.
    #
    # These methods are expected to return a list of commands to make system
    # changes when the section is loaded, deleted, or modified.
    #
    # Here is the methods are used:
    #
    # 1. When a section is loaded for the first time (new), the apply method
    # is called to perform actions such as create an interface or firewall
    # rule.
    #
    # 2. When a section is unloaded or detected as removed from a configuration
    # file, the revert method is called to undo everything that was done by
    # apply (e.g. delete an interface or rule).
    #
    # 3. When a section is reloaded (and has changed) or one or more of its
    # dependencies have have changed, the two update methods are used.
    # updateRevert selectively reverts actions that were done by apply; it can
    # (and by default does) undo everything just as revert would.  Then
    # updateApply is called to apply any changes needed to bring the system to
    # the new required state.  A motivating example is in order.
    #
    # Suppose we have one or more APs running on a single WiFi device, and we
    # loaded a new configuration that changes the channel.  The default
    # behavior would be to call revert, then call apply.  Revert would not only
    # bring down the hostapd instances but also destroy the AP-mode interfaces.
    # The latter step, however, is not necessary to achieve a channel change,
    # and if written carefully, updateRevert and updateApply can achieve the
    # configuration change with less disruption.
    #

    def apply(self, allConfigs):
        """
        Return a list of commands to apply this configuration.

        Most subclasses will need to implement this function.

        Returns a list of (priority, Command) tuples.
        """
        return []

    def revert(self, allConfigs):
        """
        Return a list of commands to revert this configuration.

        Most subclasses will need to implement this function.

        Returns a list of (priority, Command) tuples.
        """
        return []

    def updateApply(self, new, allConfigs):
        """
        Return a list of commands to update to new configuration.

        Implementing this is optional for subclasses.  The default behavior is
        to call apply.

        Returns a list of (priority, Command) tuples.
        """
        return new.apply(allConfigs)

    def updateRevert(self, new, allConfigs):
        """
        Return a list of commands to (partially) revert the configuration.

        The implementation can be selective about what it reverts (e.g. do not
        delete an interface if we are only updating its IP address).  The
        default behavior is to call revert.

        Returns a list of (priority, Command) tuples.
        """
        return self.revert(allConfigs)

    @classmethod
    def build(cls, manager, source, name, options, comment):
        """
        Build a config object instance from the UCI section.

        Arguments:
        source -- file containing this configuration section
        name -- name of the configuration section
                If None, a unique name will be generated.
        options -- dictionary of options loaded from the section
        comment -- comment string or None
        """
        obj = cls(name)
        obj.manager = manager
        obj.source = source
        obj.comment = comment

        for opdef in cls.options:
            found = False

            if opdef.type == list:
                if opdef.name in options:
                    value = options[opdef.name]
                    if not isinstance(value, list):
                        # Sometimes we expect a list but get a single value instead.
                        # Example:
                        #   ...
                        #   option network 'lan'
                        #   ...
                        # instead of
                        #   ...
                        #   list network 'lan'
                        #   ...
                        value = [value]
                    found = True
            elif opdef.type == bool:
                if opdef.name in options:
                    value = interpretBoolean(options[opdef.name])
                    found = True
            else:
                if opdef.name in options:
                    value = opdef.type(options[opdef.name])
                    found = True

            if not found:
                if opdef.required:
                    raise Exception("Missing required option {} in {}:{}:{}".format(
                        opdef.name, source, cls.typename, name))
                else:
                    value = opdef.default

            setattr(obj, opdef.name, value)

        # If this section specifies name as an option rather than in the header
        # line, update the name that will be used for comparison.
        if getattr(obj, "name", None) is not None and name is None:
            obj.internalName = obj.name

        obj.setup()

        return obj

    @classmethod
    def getModule(cls):
        """
        Get the module name (e.g. "dhcp", "wireless") for a ConfigObject class.
        """
        parts = cls.__module__.split(".")
        return parts[-1]

    @staticmethod
    def _assignPriority(config, assigned):
        """
        Recursively assign priorities to config objects based on dependencies.

        This is meant to be called by prioritizeConfigs.
        """
        if config in assigned:
            return assigned[config]

        priority = 0
        for parent in config.parents:
            pprio = ConfigObject._assignPriority(parent, assigned)
            if pprio >= priority:
                priority = pprio + 1

        assigned[config] = priority
        return priority

    @staticmethod
    def prioritizeConfigs(configs, reverse=False):
        """
        Assign priorities to config objects based on the dependency graph.

        Priority zero is assigned to all configs with no dependencies.

        priority(config1) > priority(config2) means config1 should be applied
        later than config2, and config1 should be reverted earlier than
        config2.  For configs with the same priority value, it is presumed
        that order does not matter.

        If reverse is True, the priorities are made negative so that traversing
        in increasing order gives the proper order for reverting.

        Returns a list of tuples (priority, config).  This format is suitable
        for heapq.
        """
        priorities = dict()
        for config in configs:
            ConfigObject._assignPriority(config, priorities)
        mult = -1 if reverse else 1
        return [(mult * prio, config) \
                for (config, prio) in six.iteritems(priorities)]


@attr.s(slots=True)
class ConfigOption(object):
    name = attr.ib(convert=str)
    type = attr.ib(default=str, validator=attr.validators.instance_of(type))
    required = attr.ib(convert=bool, default=False)
    default = attr.ib(default=None)
